# Note: Following stuff must be considered in a GangaRepository:
#
# * lazy loading
# * locking

from Ganga.Core.GangaRepository import GangaRepository, RepositoryError, InaccessibleObjectError
from Ganga.Utility.Plugin import PluginManagerError
import os
import os.path
import time
import errno
import copy
import threading

from Ganga.Core.GangaRepository.SessionLock import SessionLockManager

import Ganga.Utility.logging

from Ganga.Core.GangaRepository.PickleStreamer import to_file as pickle_to_file
from Ganga.Core.GangaRepository.PickleStreamer import from_file as pickle_from_file

from Ganga.Core.GangaRepository.VStreamer import to_file as xml_to_file
from Ganga.Core.GangaRepository.VStreamer import from_file as xml_from_file
from Ganga.Core.GangaRepository.VStreamer import XMLFileError

from Ganga.GPIDev.Base.Objects import Node
from Ganga.Core.GangaRepository import SubJobXMLList

from Ganga.GPIDev.Base.Proxy import isType, stripProxy, getName

logger = Ganga.Utility.logging.getLogger()

save_all_history = False

def check_app_hash(obj):
    """Writes a file safely, raises IOError on error"""

    isVerifiableApp = False
    isVerifiableAna = False

    if hasattr(obj, 'application'):
        if hasattr(obj.application, 'hash'):
            if obj.application.hash is not None:
                isVerifiableApp = True
    elif hasattr(obj, 'analysis'):
        if hasattr(obj.analysis, 'application'):
            if hasattr(obj.analysis.application, 'hash'):
                if obj.analysis.application.hash is not None:
                    isVerifiableAna = True

    if isVerifiableApp is True:
        hashable_app = stripProxy(obj.application)
    elif isVerifiableAna is True:
        hashable_app = stripProxy(obj.analysis.application)
    else:
        hashable_app = None

    if hashable_app is not None:
        if not hashable_app.calc_hash(verify=True):
            try:
                logger.warning("%s" % hashable_app)
                logger.warning('Protected attribute(s) of %s application (associated with %s #%s) changed!' % (getName(hashable_app), getName(obj), obj._registry_id))
            except AttributeError as err:
                logger.warning('Protected attribute(s) of %s application (associated with %s) changed!!!!' % (getName(hashable_app), getName(obj)))
                logger.warning("%s" % err)
            jobObj = stripProxy(hashable_app).getJobObject()
            if jobObj is not None:
                logger.warning('Job: %s is now possibly corrupt!' % jobObj.getFQID('.'))
            logger.warning('If you knowingly circumvented the protection, ignore this message (and, optionally,')
            logger.warning('re-prepare() the application). Otherwise, please file a bug report at:')
            logger.warning('https://github.com/ganga-devs/ganga/issues/')

def safe_save(fn, _obj, to_file, ignore_subs=''):
    """Try to save the XML for this object in as safe a way as possible"""

    # Add a global lock to make absolutely sure we don't have multiple threads writing files
    # See Github Issue 185
    with safe_save.lock:

        obj = stripProxy(_obj)
        check_app_hash(obj)

        # Create the dirs
        dirname = os.path.dirname(fn)
        if not os.path.exists(dirname):
            os.makedirs(dirname)

        # Prepare new data file
        new_name = fn + '.new'
        with open(new_name, "w") as tmpfile:
            to_file(obj, tmpfile, ignore_subs)

        # everything ready so create new data file and backup old one
        if os.path.exists(new_name):

            # Do we have an old one to backup?
            if os.path.exists(fn):
                os.rename(fn, fn + "~")

            os.rename(new_name, fn)

# Global lock for above function - See issue #185
safe_save.lock = threading.Lock()

def rmrf(name, count=0):
    """
    Safely recursively remove a file/folder from disk by first moving it then removing it
    calls self and will only attempt to move/remove a file 3 times before giving up
    """

    if count != 0:
        logger.debug("Trying again to remove: %s" % name)
        if count == 3:
            logger.error("Tried 3 times to remove file/folder: %s" % name)
            from Ganga.Core.exceptions import GangaException
            raise GangaException("Failed to remove file/folder: %s" % name)

    if os.path.isdir(name):

        try:
            remove_name = name + "_" + str(time.time()) + '__to_be_deleted_'
            os.rename(name, remove_name)
            #logger.debug("Move completed")
        except OSError as err:
            if err.errno != errno.ENOENT:
                logger.debug("rmrf Err: %s" % err)
                logger.debug("name: %s" % name)
                remove_name = name
                raise
            return

        for sfn in os.listdir(remove_name):
            try:
                rmrf(os.path.join(remove_name, sfn), count)
            except OSError as err:
                if err.errno == errno.EBUSY:
                    logger.debug("rmrf Remove err: %s" % err)
                    logger.debug("name: %s" % remove_name)
                    ## Sleep 2 sec and try again
                    time.sleep(2.)
                    rmrf(os.path.join(remove_name, sfn), count+1)
        try:
            os.removedirs(remove_name)
        except OSError as err:
            if err.errno == errno.ENOTEMPTY:
                rmrf(remove_name, count+1)
            elif err.errno != errno.ENOENT:
                logger.debug("%s" % err)
                raise
            return
    else:
        try:
            remove_name = name + "_" + str(time.time()) + '__to_be_deleted_'
            os.rename(name, remove_name)
        except OSError as err:
            if err.errno not in [errno.ENOENT, errno.EBUSY]:
                raise
            logger.debug("rmrf Move err: %s" % err)
            logger.debug("name: %s" % name)
            if err.errno == errno.EBUSY:
                rmrf(name, count+1)
            return

        try:
            os.remove(remove_name)
        except OSError as err:
            if err.errno != errno.ENOENT:
                logger.debug("%s" % err)
                logger.debug("name: %s" % remove_name)
                raise
            return


class GangaRepositoryLocal(GangaRepository):

    """GangaRepository Local"""

    def __init__(self, registry):
        """
        Initialize a Repository from within a Registry and keep a reference to the Registry which 'owns' it
        """
        super(GangaRepositoryLocal, self).__init__(registry)
        self.dataFileName = "data"
        self.sub_split = "subjobs"
        self.root = os.path.join(self.registry.location, "6.0", self.registry.name)
        self.lockroot = os.path.join(self.registry.location, "6.0")
        self.saved_paths = {}
        self.saved_idxpaths = {}
        self._cache_load_timestamp = {}
        self.printed_explanation = False
        self._fully_loaded = {}

    def startup(self):
        """ Starts a repository and reads in a directory structure.
        Raise RepositoryError"""
        self._load_timestamp = {}

        # New Master index to speed up loading of many, MANY files
        self._cache_load_timestamp = {}
        self._cached_cat = {}
        self._cached_cls = {}
        self._cached_obj = {}
        self._master_index_timestamp = 0

        self.known_bad_ids = []
        if "XML" in self.registry.type:
            self.to_file = xml_to_file
            self.from_file = xml_from_file
        elif "Pickle" in self.registry.type:
            self.to_file = pickle_to_file
            self.from_file = pickle_from_file
        else:
            raise RepositoryError(self.repo, "Unknown Repository type: %s" % self.registry.type)
        self.sessionlock = SessionLockManager(self, self.lockroot, self.registry.name)
        self.sessionlock.startup()
        # Load the list of files, this time be verbose and print out a summary
        # of errors
        self.update_index(verbose=True, firstRun=True)
        logger.debug("GangaRepositoryLocal Finished Startup")

    def shutdown(self):
        """Shutdown the repository. Flushing is done by the Registry
        Raise RepositoryError
        Write an index file for all new objects in memory and master index file of indexes"""
        from Ganga.Utility.logging import getLogger
        logger = getLogger()
        logger.debug("Shutting Down GangaRepositoryLocal: %s" % self.registry.name)
        for k in self._fully_loaded.keys():
            self.index_write(k, shutdown=True)
        self._write_master_cache(True)
        self.sessionlock.shutdown()

    def get_fn(self, this_id):
        """ Returns the file name where the data for this object id is saved"""
        if this_id not in self.saved_paths:
            self.saved_paths[this_id] = os.path.join(self.root, "%ixxx" % int(this_id * 0.001), "%i" % this_id, self.dataFileName)
        return self.saved_paths[this_id]

    def get_idxfn(self, this_id):
        """ Returns the file name where the data for this object id is saved"""
        if this_id not in self.saved_idxpaths:
            self.saved_idxpaths[this_id] = os.path.join(self.root, "%ixxx" % int(this_id * 0.001), "%i.index" % this_id)
        return self.saved_idxpaths[this_id]

    def index_load(self, this_id):
        """ load the index file for this object if necessary
            Loads if never loaded or timestamp changed. Creates object if necessary
            Returns True if this object has been changed, False if not
            Raise IOError on access or unpickling error 
            Raise OSError on stat error
            Raise PluginManagerError if the class name is not found"""
        #logger.debug("Loading index %s" % this_id)
        fn = self.get_idxfn(this_id)
        # index timestamp changed
        if self._cache_load_timestamp.get(this_id, 0) != os.stat(fn).st_ctime:
            logger.debug("%s != %s" % (self._cache_load_timestamp.get(this_id, 0), os.stat(fn).st_ctime))
            try:
                with open(fn, 'r') as fobj:
                    cat, cls, cache = pickle_from_file(fobj)[0]
            except Exception as x:
                logger.debug("index_load Exception: %s" % x)
                raise IOError("Error on unpickling: %s %s" %(getName(x), x))
            if this_id in self.objects:
                obj = self.objects[this_id]
                setattr(obj, "_registry_refresh", True)
            else:
                try:
                    obj = self._make_empty_object_(this_id, cat, cls)
                except Exception as err:
                    raise IOError('Failed to Parse information in Index file: %s. Err: %s' % (fn, err))
            this_cache = obj.getNodeIndexCache()
            this_data = this_cache if this_cache else {}
            for k, v in cache.iteritems():
                this_data[k] = v
            #obj.setNodeData(this_data)
            obj.setNodeIndexCache(cache)
            self._cache_load_timestamp[this_id] = os.stat(fn).st_ctime
            self._cached_cat[this_id] = cat
            self._cached_cls[this_id] = cls
            self._cached_obj[this_id] = cache
            return True
        elif this_id not in self.objects:
            self.objects[this_id] = self._make_empty_object_(this_id, self._cached_cat[this_id], self._cached_cls[this_id])
            self.objects[this_id].setNodeIndexCache( self._cached_obj[this_id] )
            setattr(self.objects[this_id], '_registry_refresh', True)
            return True
        else:
            logger.debug("Doubly loading of object with ID: %s" % this_id)
            logger.debug("Just silently continuing")
        return False

    def index_write(self, this_id, shutdown=False):
        """ write an index file for this object (must be locked).
            Should not raise any Errors,
            shutdown=True causes this to always be written regardless of any checks"""
        logger.debug("Writing index: %s" % this_id)
        obj = self.objects[this_id]
        try:
            ifn = self.get_idxfn(this_id)
            new_idx_cache = self.registry.getIndexCache(stripProxy(obj))
            if not os.path.exists(ifn) or shutdown:
                new_cache = new_idx_cache
                with open(ifn, "w") as this_file:
                    new_index = (obj._category, getName(obj), new_cache)
                    logger.debug("Writing: %s" % str(new_index))
                    pickle_to_file(new_index, this_file)
                self._cached_obj[this_id] = new_cache
                obj.setNodeIndexCache({})
            self._cached_obj[this_id] = new_idx_cache
        except IOError as err:
            logger.error("Index saving to '%s' failed: %s %s" % (ifn, getName(err), err))

    def get_index_listing(self):
        """Get dictionary of possible objects in the Repository: True means index is present,
            False if not present
        Raise RepositoryError"""
        try:
            obj_chunks = [d for d in os.listdir(self.root) if d.endswith("xxx") and d[:-3].isdigit()]
        except OSError as err:
            logger.debug("get_index_listing Exception: %s" % err)
            raise RepositoryError(self, "Could not list repository '%s'!" % (self.root))
        objs = {}  # True means index is present, False means index not present
        for c in obj_chunks:
            try:
                listing = os.listdir(os.path.join(self.root, c))
            except OSError as err:
                logger.debug("get_index_listing Exception: %s")
                raise RepositoryError(self, "Could not list repository '%s'!" % (os.path.join(self.root, c)))
            objs.update(dict([(int(l), False) for l in listing if l.isdigit()]))
            for l in listing:
                if l.endswith(".index") and l[:-6].isdigit():
                    this_id = int(l[:-6])
                    if this_id in objs:
                        objs[this_id] = True
                    else:
                        try:
                            rmrf(self.get_idxfn(this_id))
                            logger.warning("Deleted index file without data file: %s" % self.get_idxfn(this_id))
                        except OSError as err:
                            logger.debug("get_index_listing delete Exception: %s" % err)
        return objs

    def _read_master_cache(self):
        """
        read in the master cache to reduce significant I/O over many indexes separately on startup
        """
        try:
            _master_idx = os.path.join(self.root, 'master.idx')
            if os.path.isfile(_master_idx):
                logger.debug("Reading Master index")
                self._master_index_timestamp = os.stat(_master_idx).st_ctime
                with open(_master_idx, 'r') as input_f:
                    this_master_cache = pickle_from_file(input_f)[0]
                for this_cache in this_master_cache:
                    this_id = this_cache[0]
                    self._cache_load_timestamp[this_id] = this_cache[1]
                    self._cached_cat[this_id] = this_cache[2]
                    self._cached_cls[this_id] = this_cache[3]
                    self._cached_obj[this_id] = this_cache[4]
            else:
                logger.debug("Not Reading Master Index")
        except Exception as err:
            Ganga.Utility.logging.log_unknown_exception()
            logger.debug("Master Index corrupt, ignoring it")
            logger.debug("Exception: %s" % err)
            self._clear_stored_cache()

    def _clear_stored_cache(self):
        """
        clear the master cache(s) which have been stored in memory
        """
        for k, v in self._cache_load_timestamp.iteritems():
            self._cache_load_timestamp.pop(k)
        for k, v in self._cached_cat.iteritems():
            self._cached_cat.pop(k)
        for k, v in self._cached_cls.iteritems():
            self._cached_cls.pop(k)
        for k, v in self._cached_obj.iteritems():
            self._cached_obj.pop(k)

    def _write_master_cache(self, shutdown=False):
        """
        write a master index cache once per 300sec
        shutdown=True causes this to be written now
        """
        #logger.info("Updating master index: %s" % self.registry.name)
        try:
            _master_idx = os.path.join(self.root, 'master.idx')
            this_master_cache = []
            if os.path.isfile(_master_idx) and not shutdown:
                if abs(self._master_index_timestamp - os.stat(_master_idx).st_ctime) < 300:
                    return

            items_to_save = self.objects.iteritems()
            #logger.info("Updating Items: %s" % self.objects.keys())
            for k, v in items_to_save:
                try:
                    #logger.info("Examining: %s" % k)
                    if k in self._fully_loaded.keys():
                        # Check and write index first
                        obj = v#self.objects[k]
                        new_index = None
                        if obj is not None:
                            new_index = self.registry.getIndexCache(stripProxy(obj))

                        if new_index is not None:
                            #logger.debug("k: %s" % k)
                            arr_k = [k]
                            if len(self.lock(arr_k)) != 0:
                                self.index_write(k)
                                self.unlock(arr_k)
                                #stripProxy(obj).setNodeIndexCache(new_index)
                                self._cached_obj[k] = new_index

                except Exception as err:
                    logger.debug("Failed to update index: %s on startup/shutdown" % k)
                    logger.debug("Reason: %s" % err)

            #cached_list = []
            iterables = self._cache_load_timestamp.iteritems()
            for k, v in iterables:
                cached_list = []
                cached_list.append(k)
                try:
                    fn = self.get_idxfn(k)
                    time = os.stat(fn).st_ctime
                except OSError as err:
                    logger.debug("_write_master_cache: %s" % err)
                    logger.debug("_cache_load_timestamp: %s" % self._cache_load_timestamp)
                    import errno
                    if err.errno == errno.ENOENT:  # If file is not found
                        time = -1
                    else:
                        raise
                if time > 0:
                    cached_list.append(time)
                    cached_list.append(self._cached_cat[k])
                    cached_list.append(self._cached_cls[k])
                    cached_list.append(self._cached_obj[k])
                    this_master_cache.append(cached_list)

            try:
                with open(_master_idx, 'w') as of:
                    pickle_to_file(this_master_cache, of)
            except IOError as err:
                logger.debug("write_master: %s" % err)
                try:
                    os.remove(os.path.join(self.root, 'master.idx'))
                except OSError as x:
                    Ganga.Utility.logging.log_user_exception(debug=True)
        except Exception as err:
            logger.debug("write_error2: %s" % err)
            Ganga.Utility.logging.log_unknown_exception()

        return

    def updateLocksNow(self):
        """
        Trigger the session locks to all be updated now
        This is useful when the SessionLock is updating either too slowly or has gone to sleep when there are multiple sessions
        """
        self.sessionlock.updateNow()

    def update_index(self, this_id=None, verbose=False, firstRun=False):
        """ Update the list of available objects
        Raise RepositoryError"""
        # First locate and load the index files
        logger.debug("updating index...")
        objs = self.get_index_listing()
        changed_ids = []
        deleted_ids = set(self.objects.keys())
        summary = []
        if firstRun:
            self._read_master_cache()
        logger.debug("Iterating over Items")

        locked_ids = self.sessionlock.locked

        for this_id in objs.keys():
            deleted_ids.discard(this_id)
            # Make sure we do not overwrite older jobs if someone deleted the
            # count file
            if this_id > self.sessionlock.count:
                self.sessionlock.count = this_id + 1
            # Locked IDs can be ignored
            if this_id in locked_ids:
                continue
            # Now we treat unlocked IDs
            try:
                # if this succeeds, all is well and we are done
                if self.index_load(this_id):
                    changed_ids.append(this_id)
                continue
            except IOError as err:
                logger.debug("IOError: Failed to load index %i: %s" % (this_id, err))
            except OSError as err:
                logger.debug("OSError: Failed to load index %i: %s" % (this_id, err))
            except PluginManagerError as err:
                # Probably should be DEBUG
                logger.debug("PluginManagerError: Failed to load index %i: %s" % (this_id, err))
                # This is a FATAL error - do not try to load the main file, it
                # will fail as well
                summary.append((this_id, err))
                continue

            # this is bad - no or corrupted index but object not loaded yet!
            # Try to load it!
            if not this_id in self.objects:
                try:
                    self.load([this_id])
                    changed_ids.append(this_id)
                    # Write out a new index if the file can be locked
                    if len(self.lock([this_id])) != 0:
                        self.index_write(this_id)
                        self.unlock([this_id])
                except KeyError as err:
                    logger.debug("update Error: %s" % err)
                    # deleted job
                    if this_id in self.objects:
                        self._internal_del__(this_id)
                        changed_ids.append(this_id)
                except Exception as x:
                    ## WE DO NOT CARE what type of error occured here and it can be
                    ## due to corruption so could be one of MANY exception types
                    ## If the job is not accessible this should NOT cause the loading of ganga to fail!
                    ## we can't reasonably write all possible exceptions here!
                    logger.debug("update_index: Failed to load id %i: %s" % (this_id, x))
                    summary.append((this_id, x))
                #finally:
                #    pass

        logger.debug("Iterated over Items")

        # Check deleted files:
        for this_id in deleted_ids:
            self._internal_del__(this_id)
            changed_ids.append(this_id)
        if len(deleted_ids) > 0:
            logger.warning("Registry '%s': Job %s externally deleted." % (self.registry.name, ",".join(map(str, list(deleted_ids)))))

        if len(summary) > 0:
            cnt = {}
            examples = {}
            for this_id, x in summary:
                if this_id in self.known_bad_ids:
                    continue
                cnt[getName(x)] = cnt.get(getName(x), []) + [str(this_id)]
                examples[getName(x)] = str(x)
                self.known_bad_ids.append(this_id)
                # add object to incomplete_objects
                if not this_id in self.incomplete_objects:
                    self.incomplete_objects.append(this_id)

            for exc, ids in cnt.items():
                logger.error("Registry '%s': Failed to load %i jobs (IDs: %s) due to '%s' (first error: %s)" % (self.registry.name, len(ids), ",".join(ids), exc, examples[exc]))

            if self.printed_explanation is False:
                logger.error("If you want to delete the incomplete objects, you can type:\n")
                logger.error("'for i in %s.incomplete_ids(): %s(i).remove()'\n (then press 'Enter' twice)" % (self.registry.name, self.registry.name))
                logger.error("WARNING!!! This will result in corrupt jobs being completely deleted!!!")
                self.printed_explanation = True
        logger.debug("updated index done")

        if len(changed_ids) != 0:
            isShutdown = not firstRun
            self._write_master_cache(shutdown=isShutdown)

        return changed_ids

    def add(self, objs, force_ids=None):
        """ Add the given objects to the repository, forcing the IDs if told to.
        Raise RepositoryError"""

        logger.debug("add")

        if force_ids not in [None, []]:  # assume the ids are already locked by Registry
            if not len(objs) == len(force_ids):
                raise RepositoryError(self, "Internal Error: add with different number of objects and force_ids!")
            ids = force_ids
        else:
            ids = self.sessionlock.make_new_ids(len(objs))

        logger.debug("made ids")

        for i in range(0, len(objs)):
            fn = self.get_fn(ids[i])
            try:
                os.makedirs(os.path.dirname(fn))
            except OSError as e:
                if e.errno != errno.EEXIST:
                    raise RepositoryError( self, "OSError on mkdir: %s" % (e))
            self._internal_setitem__(ids[i], objs[i])

            # Set subjobs dirty - they will not be flushed if they are not.
            if self.sub_split and self.sub_split in objs[i].getNodeData():
                try:
                    sj_len = len(objs[i].getNodeAttribute(self.sub_split))
                    if sj_len > 0:
                        for j in range(sj_len):
                            objs[i].getNodeAttribute(self.sub_split)[j]._dirty = True
                except AttributeError as err:
                    logger.debug("RepoXML add Exception: %s" % err)

        logger.debug("Added")

        return ids

    def _safe_flush_xml(self, this_id):
        """
        Flush XML to disk whilst checking for relavent SubJobXMLList which handles subjobs now
        flush for "this_id" in the self.objects list
        """

        fn = self.get_fn(this_id)
        obj = self.objects[this_id]
        from Ganga.Core.GangaRepository.VStreamer import EmptyGangaObject
        if not isType(obj, EmptyGangaObject):
            split_cache = None

            has_children = (not self.sub_split is None) and\
                    (self.sub_split in obj.getNodeData()) and obj.getNodeAttribute(self.sub_split) and len(obj.getNodeAttribute(self.sub_split)) > 0

            if has_children:

                logger.debug("has_children")

                if hasattr(obj.getNodeAttribute(self.sub_split), 'flush'):
                    # I've been read from disk in the new SubJobXMLList format I know how to flush
                    obj.getNodeAttribute(self.sub_split).flush()
                else:
                    # I have been constructed in this session, I don't know how to flush!
                    if hasattr(obj.getNodeAttribute(self.sub_split)[0], "_dirty"):
                        split_cache = obj.getNodeAttribute(self.sub_split)
                        for i in range(len(split_cache)):
                            if not split_cache[i]._dirty:
                                continue
                            sfn = os.path.join(os.path.dirname(fn), str(i), self.dataFileName)
                            if not os.path.exists(os.path.dirname(sfn)):
                                logger.debug("Constructing Folder: %s" % os.path.dirname(sfn))
                                os.makedirs(os.path.dirname(sfn))
                            else:
                                logger.debug("Using Folder: %s" % os.path.dirname(sfn))
                            safe_save(sfn, split_cache[i], self.to_file)
                            split_cache[i]._setFlushed()
                    from Ganga.Core.GangaRepository.SubJobXMLList import SubJobXMLList
                    # Now generate an index file to take advantage of future non-loading goodness
                    tempSubJList = SubJobXMLList(os.path.dirname(fn), self.registry, self.dataFileName, False, parent=obj)
                    ## equivalent to for sj in job.subjobs
                    tempSubJList._setParent(obj)
                    job_dict = {}
                    for sj in obj.getNodeAttribute(self.sub_split):
                        job_dict[sj.id] = stripProxy(sj)
                    tempSubJList._reset_cachedJobs(job_dict)
                    tempSubJList.flush()
                    del tempSubJList

                safe_save(fn, obj, self.to_file, self.sub_split)
                # clean files not in subjobs anymore... (bug 64041)
                for idn in os.listdir(os.path.dirname(fn)):
                    split_cache = obj.getNodeAttribute(self.sub_split)
                    if idn.isdigit() and int(idn) >= len(split_cache):
                        rmrf(os.path.join(os.path.dirname(fn), idn))
            else:

                logger.debug("not has_children")

                safe_save(fn, obj, self.to_file, "")
                # clean files leftover from sub_split
                for idn in os.listdir(os.path.dirname(fn)):
                    if idn.isdigit():
                        rmrf(os.path.join(os.path.dirname(fn), idn))
            self.index_write(this_id)
            #obj.setNodeIndexCache(None)
            obj._setFlushed()
        else:
            raise RepositoryError(self, "Cannot flush an Empty object for ID: %s" % this_id)

        if this_id not in self._fully_loaded.keys():
            self._fully_loaded[this_id] = obj

    def flush(self, ids):
        """
        flush the set of "ids" to disk and write the XML representing said objects in self.objects
        """
        logger.debug("Flushing: %s" % ids)

        #import traceback
        #traceback.print_stack()
        for this_id in ids:
            try:
                logger.debug("safe_flush")
                self._safe_flush_xml(this_id)

                self._cache_load_timestamp[this_id] = time.time()
                self._cached_cls[this_id] = getName(self.objects[this_id])
                self._cached_cat[this_id] = self.objects[this_id]._category
                self._cached_obj[this_id] = self.objects[this_id].getNodeIndexCache()

                try:
                    self.index_write(this_id)
                except:
                    logger.debug("Index write failed")
                    pass

                if this_id not in self._fully_loaded.keys():
                    self._fully_loaded[this_id] = self.objects[this_id]

            except (OSError, IOError, XMLFileError) as x:
                raise RepositoryError(self, "Error of type: %s on flushing id '%s': %s" % (type(x), this_id, x))

    def is_loaded(self, this_id):
        """
        Has "this_id" been loaded from disk already
        Check self.objects for the ._data sttribute
        """
        return (this_id in self.objects) and (self.objects[this_id].getNodeData() is not None)

    def count_nodes(self, this_id):
        """
        Count the number of nodes in "this_id" of self.objects"
        Is this used any more?
        """
        node_count = 0
        fn = self.get_fn(this_id)

        ld = os.listdir(os.path.dirname(fn))
        i = 0
        while str(i) in ld:
            sfn = os.path.join(os.path.dirname(fn), str(i), self.dataFileName)
            if os.path.exists(sfn):
                node_count = node_count + 1
            i += 1

        return node_count

    def _actually_loaded(self, this_id):
        """
        Return if "this_id" is in the _full_loaded dictionary of objects loaded from XML by this class
        This is different to is_loaded as it checks for actual read operations
        """
        return this_id in self._fully_loaded.keys()

    def _check_index_cache(self, obj, this_id):
        """
        Checks the index cache of "this_id" against the index cache generated from the "obj"ect
        If there is a problem then the object is unloaded from memory but will not do anything if everything agrees here
        """
        new_idx_cache = self.registry.getIndexCache(stripProxy(obj))
        if new_idx_cache != obj.getNodeIndexCache():
            logger.debug("NEW: %s" % new_idx_cache)
            logger.debug("OLD: %s" % obj.getNodeIndexCache())
            # index is wrong! Try to get read access - then we can fix this
            if len(self.lock([this_id])) != 0:
                self.index_write(this_id)
                # self.unlock([this_id])

                old_idx_subset = all((k in new_idx_cache and new_idx_cache[k] == v) for k, v in obj.getNodeIndexCache().iteritems())
                if not old_idx_subset:
                    # Old index cache isn't subset of new index cache
                    new_idx_subset = all((k in obj.getNodeIndexCache() and obj.getNodeIndexCache()[k] == v) for k, v in new_idx_cache.iteritems())
                else:
                    # Old index cache is subset of new index cache so no need to check
                    new_idx_subset = True

                if not old_idx_subset and not new_idx_subset:
                    logger.warning("Incorrect index cache of '%s' object #%s was corrected!" % (self.registry.name, this_id))
                    logger.debug("old cache: %s\t\tnew cache: %s" % (obj.getNodeIndexCache(), new_idx_cache))
                    self.unlock([this_id])
            else:
                pass
                # if we cannot lock this, the inconsistency is
                # most likely the result of another ganga
                # process modifying the repo
                #obj.setNodeIndexCache(None)

    def _must_actually_load_xml(self, fn, this_id, load_backup, has_children, tmpobj):
        """
        If we must actually load the object from disk then we end up here.
        This replaces the attrs of "objects[this_id]" with the attrs from tmpobj
        If there are children then a SubJobXMLList is created to manage them.
        The fn of the job is passed to the SubbJobXMLList and there is some knowledge of if we should be loading the backup passed as well
        """
        obj = self.objects[this_id]
        for key, val in tmpobj.getNodeData().iteritems():
            obj.setNodeAttribute(key, val)
        for attr_name, attr_val in obj._schema.allItems():
            if attr_name not in tmpobj.getNodeData().keys():
                obj.setNodeAttribute(attr_name, obj._schema.getDefaultValue(attr_name))

        if has_children:
        #    logger.info("Adding children")
            obj.setNodeAttribute(self.sub_split, SubJobXMLList.SubJobXMLList(os.path.dirname(fn), self.registry, self.dataFileName, load_backup, parent=obj))
        else:
            obj.setNodeAttribute(self.sub_split, None)

        from Ganga.GPIDev.Base.Objects import do_not_copy
        for node_key, node_val in obj.getNodeData().iteritems():
            if isType(node_val, Node):
                if node_key not in do_not_copy:
                    node_val._setParent(obj)

        # Check if index cache; if loaded; was valid:
        if obj.getNodeIndexCache() not in [{}]:
            self._check_index_cache(obj, this_id)

        obj.setNodeIndexCache({})

        if this_id not in self._fully_loaded.keys():
            self._fully_loaded[this_id] = obj

    def _actually_load_xml(self, fobj, fn, this_id, load_backup):
        """
        This is the method which will load the job from fn using the fobj using the self.from_file method and _must_actually_load_xml is called to replace the
        self.objects[this_id] with the correct attributes. We also preseve knowledge of if we're being asked to load a backup or not
        """
        #print("ACTUALLY LOAD")

        tmpobj = None

        if (self._load_timestamp.get(this_id, 0) != os.fstat(fobj.fileno()).st_ctime):

            import time
            b4=time.time()
            tmpobj, errs = self.from_file(fobj)
            a4=time.time()
            logger.debug("Loading XML file for ID: %s took %s sec" % (this_id, a4-b4))

            if len(errs) > 0:
                logger.error("#%s Error(s) Loading File: %s" % (len(errs), fobj.name))
                raise InaccessibleObjectError(self, this_id, errs[0])

            has_children = (self.sub_split is not None) and (self.sub_split in tmpobj.getNodeData()) and len(tmpobj.getNodeAttribute(self.sub_split)) == 0

            if this_id in self.objects:
                self._must_actually_load_xml(fn, this_id, load_backup, has_children, tmpobj)

            else:
                #tmpobj.setNodeIndexCache(None)
                self._internal_setitem__(this_id, tmpobj)

            if hasattr(self.objects[this_id], self.sub_split):
                sub_attr = getattr(self.objects[this_id], self.sub_split)
                if sub_attr is not None and hasattr(sub_attr, '_setParent'):
                    sub_attr._setParent(self.objects[this_id])

            self._load_timestamp[this_id] = os.fstat(fobj.fileno()).st_ctime

        else:
            logger.debug("Didn't Load Job ID: %s" % this_id)

        logger.debug("Finished Loading XML")

    def _open_xml_file(self, fn, this_id, _copy_backup=False):
        """
        This loads the XML for the job "this_id" in self.objects using the file "fn" and knowing whether we want the file or the backup by _copy_backup
        """
        fobj = None

        has_loaded_backup = False

        try:
            if not os.path.isfile(fn) and _copy_backup:
                if os.path.isfile(fn + '~'):
                    logger.warning("XML File: %s missing, recovering from backup, recent changes may have been lost!" % fn)
                    has_loaded_backup = True
                    try:
                        from shutil import copyfile
                        copyfile(fn+'~', fn)
                    except:
                        logger.warning("Error Recovering the backup file! loading of Job may Fail!")
            fobj = open(fn, "r")
        except IOError as x:
            if x.errno == errno.ENOENT:
                # remove index so we do not continue working with wrong information
                try:
                    # remove internal representation
                    self._internal_del__(this_id)
                    rmrf(os.path.dirname(fn) + ".index")
                except OSError as err:
                    logger.debug("load unlink Error: %s" % err)
                    pass
                raise KeyError(this_id)
            else:
                raise RepositoryError(self, "IOError: %s" % x)
        finally:
            try:
                if os.path.isdir(os.path.dirname(fn)):
                    ld = os.listdir(os.path.dirname(fn))
                    if len(ld) == 0:
                        os.rmdir(os.path.dirname(fn))
                        logger.warning("No job index or data found, removing empty directory: %s" % os.path.dirname(fn))
            except Exception as err:
                logger.debug("load error %s" % err)
                pass

        return fobj, has_loaded_backup

    def load(self, ids, load_backup=False):
        """
        Load the following "ids" from disk
        If we want to load the backup files for these ids then use _copy_backup
        Correctly loaded objects are dirty, Objects loaded from backups for whatever reason are marked dirty
        """
        #print("load: %s " % ids)
        #import traceback
        #traceback.print_stack()
        #print("\n")

        logger.debug("Loading Repo object(s): %s" % ids)

        for this_id in ids:

            fn = self.get_fn(this_id)
            if load_backup:
                has_loaded_backup = True
                fn = fn + "~"
            else:
                has_loaded_backup = False

            fobj = None

            try:
                fobj, has_loaded_backup2 = self._open_xml_file(fn, this_id, _copy_backup=True)
                if has_loaded_backup2:
                    has_loaded_backup = has_loaded_backup2
            except Exception as err:
                logger.debug("XML load: Failed to load XML file: %s" % fn)
                logger.debug("Error was:\n%s" % err)
                raise

            try:
                self._actually_load_xml(fobj, fn, this_id, load_backup)
            except RepositoryError as err:
                logger.debug("Repo Exception: %s" % err)
                raise

            except Exception as err:

                should_continue = self._handle_load_exception(err, fn, this_id, load_backup)

                if should_continue is True:
                    has_loaded_backup = True
                    continue
                else:
                    raise

            finally:
                fobj.close()

            if has_loaded_backup:
                self.objects[this_id]._setDirty()
            else:
                self.objects[this_id]._setFlushed()

        logger.debug("Finished 'load'-ing of: %s" % ids)


    def _handle_load_exception(self, err, fn, this_id, load_backup):
        """
        This method does a lot of the handling of an exception thrown from the load method
        We will return True/False here, True if the error can be correctly caught and False if this is terminal and we couldn't load the object
        """
        if isType(err, XMLFileError):
            logger.error("XML File failed to load for Job id: %s" % this_id)
            logger.error("Actual Error was:\n%s" % err)

        if load_backup:
            logger.debug("Could not load backup object #%s: %s" % (this_id, err))
            raise InaccessibleObjectError(self, this_id, err)

        logger.debug("Could not load object #%s: %s" % (this_id, err))

        # try loading backup
        try:
            self.load([this_id], load_backup=True)
            logger.warning("Object '%s' #%s loaded from backup file - recent changes may be lost." % (self.registry.name, this_id))
            return True
        except Exception as err2:
            logger.debug("Exception when loading backup: %s" % err2 )
        #finally:
        #    pass

        logger.error("XML File failed to load for Job id: %s" % this_id)
        logger.error("Actual Error was:\n%s" % err2)

        # add object to incomplete_objects
        if not this_id in self.incomplete_objects:
            self.incomplete_objects.append(this_id)
            # remove index so we do not continue working with wrong
            # information
            rmrf(os.path.dirname(fn) + ".index")
            raise InaccessibleObjectError(self, this_id, err)

        return False

    def delete(self, ids):
        """
        This is the method to 'delete' an object from disk, it's written in python and starts with the indexes first
        """
        for this_id in ids:
            # First remove the index, so that it is gone if we later have a
            # KeyError
            fn = self.get_fn(this_id)
            try:
                rmrf(os.path.dirname(fn) + ".index")
            except OSError as err:
                logger.debug("Delete Error: %s" % err)
            self._internal_del__(this_id)
            rmrf(os.path.dirname(fn))
            if this_id in self._fully_loaded.keys():
                del self._fully_loaded[this_id]
            if this_id in self.objects:
                del self.objects[this_id]

    def lock(self, ids):
        """
        Request a session lock for the following ids
        """
        return self.sessionlock.lock_ids(ids)

    def unlock(self, ids):
        """
        Unlock (release file locks of) the following ids
        """
        released_ids = self.sessionlock.release_ids(ids)
        if len(released_ids) < len(ids):
            logger.error("The write locks of some objects could not be released!")

    def get_lock_session(self, this_id):
        """get_lock_session(id)
        Tries to determine the session that holds the lock on id for information purposes, and return an informative string.
        Returns None on failure
        """
        return self.sessionlock.get_lock_session(this_id)

    def get_other_sessions(self):
        """get_session_list()
        Tries to determine the other sessions that are active and returns an informative string for each of them.
        """
        return self.sessionlock.get_other_sessions()

    def reap_locks(self):
        """reap_locks() --> True/False
        Remotely clear all foreign locks from the session.
        WARNING: This is not nice.
        Returns True on success, False on error."""
        return self.sessionlock.reap_locks()

    def clean(self):
        """clean() --> True/False
        Clear EVERYTHING in this repository, counter, all jobs, etc.
        WARNING: This is not nice."""
        self.shutdown()
        rmrf(self.root)
        self.startup()

    def updateIndexCache(self, obj):
        """
        This method should be obsoleted fairly soon as it's no longer needed. We now handle caches correctly
        """
        #stripProxy(obj).setNodeIndexCache(self.registry.getIndexCache(stripProxy(obj)))
        pass




from __future__ import absolute_import
from __future__ import print_function
from Ganga.GPIDev.Base import GangaObject
from Ganga.GPIDev.Schema import Schema, Version, SimpleItem, ComponentItem, FileItem, GangaFileItem
from Ganga.Utility.logging import getLogger
from Ganga.Utility.ColourText import status_colours, overview_colours, ANSIMarkup
markup = ANSIMarkup()
from Ganga.GPIDev.Lib.Tasks.common import getJobByID
from Ganga.Core.exceptions import ApplicationConfigurationError
from Ganga.GPIDev.Base.Proxy import stripProxy
import time
from Ganga.GPIDev.Lib.Tasks.ITask import addInfoString
import sys
import traceback
from Ganga.GPIDev.Adapters.IGangaFile import IGangaFile
from Ganga.GPIDev.Lib.File.File import File

logger = getLogger()

def formatTraceback():
   "Helper function to printout a traceback as a string"
   return "\n %s\n%s\n%s\n" % (''.join( traceback.format_tb(sys.exc_info()[2])), sys.exc_info()[0], sys.exc_info()[1])

class IUnit(GangaObject):
    _schema = Schema(Version(1, 0), {
        'status': SimpleItem(defvalue='new', protected=1, copyable=0, doc='Status - running, pause or completed', typelist=[str]),
        'name': SimpleItem(defvalue='Simple Unit', doc='Name of the unit (cosmetic)', typelist=[str]),
        'application': ComponentItem('applications', defvalue=None, optional=1, load_default=False, doc='Application of the Transform.'),
        'inputdata': ComponentItem('datasets', defvalue=None, optional=1, load_default=False, doc='Input dataset'),
        'outputdata': ComponentItem('datasets', defvalue=None, optional=1, load_default=False, doc='Output dataset'),
        'active': SimpleItem(defvalue=False, hidden=1, doc='Is this unit active'),
        'active_job_ids': SimpleItem(defvalue=[], typelist=[int], sequence=1, hidden=1, doc='Active job ids associated with this unit'),
        'prev_job_ids': SimpleItem(defvalue=[], typelist=[int], sequence=1,  hidden=1, doc='Previous job ids associated with this unit'),
        'minor_resub_count': SimpleItem(defvalue=0, hidden=1, doc='Number of minor resubmits'),
        'major_resub_count': SimpleItem(defvalue=0, hidden=1, doc='Number of major resubmits'),
        'req_units': SimpleItem(defvalue=[], typelist=[str], sequence=1, hidden=1, doc='List of units that must complete for this to start (format TRF_ID:UNIT_ID)'),
        'start_time': SimpleItem(defvalue=0, hidden=1, doc='Start time for this unit. Allows a delay to be put in'),
        'copy_output': ComponentItem('datasets', defvalue=None, load_default=0, optional=1, doc='The dataset to copy the output of this unit to, e.g. Grid dataset -> Local Dataset'),
        'merger': ComponentItem('mergers', defvalue=None, load_default=0, optional=1, doc='Merger to be run after this unit completes.'),
        'splitter': ComponentItem('splitters', defvalue=None, optional=1, load_default=False, doc='Splitter used on each unit of the Transform.'),
        'postprocessors': ComponentItem('postprocessor', defvalue=None, doc='list of postprocessors to run after job has finished'),
        'inputsandbox': FileItem(defvalue=[], sequence=1, doc="list of File objects shipped to the worker node "),
        'inputfiles': GangaFileItem(defvalue=[], sequence=1, doc="list of file objects that will act as input files for a job"),
        'outputfiles': GangaFileItem(defvalue=[], sequence=1, doc="list of OutputFile objects to be copied to all jobs"),
        'info' : SimpleItem(defvalue=[],typelist=[str],protected=1,sequence=1,doc="Info showing status transitions and unit info"),
        'id': SimpleItem(defvalue=-1, protected=1, doc='ID of the Unit', typelist=[int]),
    })

    _category = 'units'
    _name = 'IUnit'
    _exportmethods = []
    _hidden = 0

# Special methods:
    def __init__(self):
        super(IUnit, self).__init__()
        self.updateStatus("new")

    def _readonly(self):
        """A unit is read-only if the status is not new."""
        if self.status == "new":
            return 0
        return 1

    def validate(self):
        """Validate that this unit is OK and set it to active"""
        self.active = True
        return True

    def getID(self):
        """Get the ID of this unit within the transform"""

        # if the id isn't already set, use the index from the parent Task
        if self.id < 0:
           trf = self._getParent()
           if not trf:
              raise ApplicationConfigurationError(
                 None, "This unit has not been associated with a transform and so there is no ID available")
           self.id = trf.units.index(self)
           
        return self.id

    def updateStatus(self, status):
        """Update status hook"""
        addInfoString(self, "Status change from '%s' to '%s'" % (self.status, status))
        self.status = status

    def createNewJob(self):
        """Create any jobs required for this unit"""
        pass

    def checkCompleted(self, job):
        """Check if this unit is complete"""
        if job.status == "completed":
            return True
        else:
            return False

    def checkForSubmission(self):
        """Check if this unit should submit a job"""

        # check the delay
        if time.time() < self.start_time:
            return False

        # check if we already have a job
        if len(self.active_job_ids) != 0:
            return False

        # if we're using threads, check the max number
        from Ganga.Core.GangaThread.WorkerThreads import getQueues
        if self._getParent().submit_with_threads and getQueues().totalNumUserThreads() > self._getParent().max_active_threads:
            return False

        return True

    def checkForResubmission(self):
        """check if this unit should be resubmitted"""

        # check if we already have a job
        if len(self.active_job_ids) == 0:
            return False
        else:
            job = getJobByID(self.active_job_ids[0])
            if job.status in ["failed", "killed"]:
                return True

            return False

    def checkParentUnitsAreComplete(self):
        """Check to see if the parent units are complete"""
        req_ok = True
        task = self._getParent()._getParent()
        for req in self.req_units:
            req_trf_id = int(req.split(":")[0])

            if req.find("ALL") == -1:
                req_unit_id = int(req.split(":")[1])
                if task.transforms[req_trf_id].units[req_unit_id].status != "completed":
                    req_ok = False

            else:
                # need all units from this trf
                for u in task.transforms[req_trf_id].units:
                    if u.status != "completed":
                        req_ok = False

        return req_ok

    def checkMajorResubmit(self, job):
        """check if this job needs to be fully rebrokered or not"""
        pass

    def majorResubmit(self, job):
        """perform a mjor resubmit/rebroker"""
        self.prev_job_ids.append(job.id)
        self.active_job_ids.remove(job.id)

    def minorResubmit(self, job):
        """perform just a minor resubmit"""
        try:
            trf = self._getParent()
        except Exception as err:
            logger.debug("GetParent exception!\n%s" % str(err))
            trf = None
        if trf is not None and trf.submit_with_threads:
            addInfoString( self, "Attempting job re-submission with queues..." )
            from Ganga.Core.GangaThread.WorkerThreads import getQueues
            getQueues().add(job.resubmit)
        else:
            addInfoString( self, "Attempting job re-submission..." )
            job.resubmit()

    def update(self):
        """Update the unit and (re)submit jobs as required"""

        # if we're complete, then just return
        if self.status in ["completed", "recreating"] or not self.active:
            return 0

        # check if submission is needed
        task = self._getParent()._getParent()
        trf = self._getParent()
        maxsub = task.n_tosub()

        # check parent unit(s)
        req_ok = self.checkParentUnitsAreComplete()

        # set the start time if not already set
        if len(self.req_units) > 0 and req_ok and self.start_time == 0:
            self.start_time = time.time() + trf.chain_delay * 60 - 1

        if req_ok and self.checkForSubmission() and maxsub > 0:

            # create job and submit
            addInfoString( self, "Creating Job..." )
            j = self.createNewJob()
            if j.name == '':
                j.name = "T%i:%i U%i" % (task.id, trf.getID(), self.getID())

            try:
                if trf.submit_with_threads:
                    addInfoString( self, "Attempting job submission with queues..." )
                    from Ganga.Core.GangaThread.WorkerThreads import getQueues
                    getQueues().add(j.submit)
                else:
                    addInfoString( self, "Attempting job submission..." )
                    j.submit()

            except Exception as err:
                logger.debug("update Err: %s" % str(err))
                addInfoString( self, "Failed Job Submission")
                addInfoString( self, "Reason: %s" % (formatTraceback()))
                logger.error("Couldn't submit the job. Deactivating unit.")
                self.prev_job_ids.append(j.id)
                self.active = False
                trf._setDirty()  # ensure everything's saved
                return 1

            self.active_job_ids.append(j.id)
            self.updateStatus("running")
            trf._setDirty()  # ensure everything's saved

            if trf.submit_with_threads:
                return 0

            return 1

        # update any active jobs
        for jid in self.active_job_ids:

            # we have an active job so see if this job is OK and resubmit if
            # not
            try:
                job = getJobByID(jid)
            except Exception as err:
                logger.debug("Update2 Err: %s" % str(err))
                logger.warning("Cannot find job with id %d. Maybe reset this unit with: tasks(%d).transforms[%d].resetUnit(%d)" %
                               (jid, task.id, trf.getID(), self.getID()))
                continue

            if job.status == "completed":

                # check if actually completed
                if not self.checkCompleted(job):
                    return 0

                # check for DS copy
                if trf.unit_copy_output:
                    if not self.copy_output:
                        trf.createUnitCopyOutputDS(self.getID())

                    if not self.copyOutput():
                        return 0

                # check for merger
                if trf.unit_merger:
                    if not self.merger:
                        self.merger = trf.createUnitMerger(self.getID())

                    if not self.merge():
                        return 0

                # all good so mark unit as completed
                self.updateStatus("completed")

            elif job.status == "failed" or job.status == "killed":

                # check for too many resubs
                if self.minor_resub_count + self.major_resub_count > trf.run_limit - 1:
                    logger.error("Too many resubmits (%i). Deactivating unit." % (
                        self.minor_resub_count + self.major_resub_count))
                    addInfoString( self, "Deactivating unit. Too many resubmits (%i)" % ( self.minor_resub_count + self.major_resub_count))
                    self.active = False
                    return 0

                rebroker = False

                if self.minor_resub_count > trf.minor_run_limit - 1:
                    if self._getParent().rebroker_on_job_fail:
                        rebroker = True
                    else:
                        logger.error(
                            "Too many minor resubmits (%i). Deactivating unit." % self.minor_resub_count)
                        addInfoString( self, "Deactivating unit. Too many resubmits (%i)" % (self.minor_resub_count + self.minor_resub_count))
                        self.active = False
                        return 0

                if self.major_resub_count > trf.major_run_limit - 1:
                    logger.error(
                        "Too many major resubmits (%i). Deactivating unit." % self.major_resub_count)
                    addInfoString( self, "Deactivating unit. Too many resubmits (%i)" % (self.minor_resub_count + self.major_resub_count))
                    self.active = False
                    return 0

                # check the type of resubmit
                if rebroker or self.checkMajorResubmit(job):

                    self.major_resub_count += 1
                    self.minor_resub_count = 0

                    try:
                        addInfoString( self, "Attempting major resubmit...")
                        self.majorResubmit(job)
                    except Exception as err:
                        logger.debug("Update Err3: %s" % str(err))
                        logger.error("Couldn't resubmit the job. Deactivating unit.")
                        addInfoString( self, "Failed Job resubmission")
                        addInfoString( self, "Reason: %s" % (formatTraceback()))
                        self.active = False

                    # break the loop now because we've probably changed the
                    # active jobs list
                    return 1
                else:
                    self.minor_resub_count += 1
                    try:
                        addInfoString( self, "Attempting minor resubmit...")
                        self.minorResubmit(job)
                    except Exception as err:
                        logger.debug("Update Err4: %s" % str(err))
                        logger.error("Couldn't resubmit the job. Deactivating unit.")
                        addInfoString( self, "Failed Job resubmission")
                        addInfoString( self, "Reason: %s" % (formatTraceback()))
                        self.active = False
                        return 1

    def reset(self):
        """Reset the unit completely"""
        addInfoString( self, "Reseting Unit...")
        self.minor_resub_count = 0
        self.major_resub_count = 0
        if len(self.active_job_ids) > 0:
            self.prev_job_ids += self.active_job_ids
        self.active_job_ids = []

        self.active = True

        # if has parents, set to recreate
        if len(self.req_units) > 0:
            self.updateStatus("recreating")
        else:
            self.updateStatus("running")

    # Info routines
    def n_active(self):

        if self.status == 'completed':
            return 0

        tot_active = 0
        active_states = ['submitted', 'running']

        for jid in self.active_job_ids:

            try:
                job = getJobByID(jid)
            except Exception as err:
                logger.debug("n_active Err: %s" % str(err))
                task = self._getParent()._getParent()
                trf = self._getParent()
                logger.warning("Cannot find job with id %d. Maybe reset this unit with: tasks(%d).transforms[%d].resetUnit(%d)" %
                               (jid, task.id, trf.getID(), self.getID()))
                continue

            j = stripProxy(job)

            # try to preserve lazy loading
            if hasattr(j, 'getNodeIndexCache') and j.getNodeIndexCache() and 'subjobs:status' in j.getNodeIndexCache():
                if len(j.getNodeIndexCache()['subjobs:status']) > 0:
                    for sj_stat in j.getNodeIndexCache()['subjobs:status']:
                        if sj_stat in active_states:
                            tot_active += 1
                else:
                    if j.getNodeIndexCache()['status'] in active_states:
                        tot_active += 1
            else:
                #logger.warning("WARNING: (active check) No index cache for job object %d" % jid)
                if j.status in active_states:
                    if j.subjobs:
                        for sj in j.subjobs:
                            if sj.status in active_states:
                                tot_active += 1
                    else:
                        tot_active += 1

        return tot_active

    def n_status(self, status):
        tot_active = 0
        for jid in self.active_job_ids:

            try:
                job = getJobByID(jid)
            except Exception as err:
                logger.debug("n_status Err: %s" % str(err))
                task = self._getParent()._getParent()
                trf = self._getParent()
                logger.warning("Cannot find job with id %d. Maybe reset this unit with: tasks(%d).transforms[%d].resetUnit(%d)" %
                               (jid, task.id, trf.getID(), self.getID()))
                continue

            j = stripProxy(job)

            # try to preserve lazy loading
            if hasattr(j, 'getNodeIndexCache') and j.getNodeIndexCache() and 'subjobs:status' in j.getNodeIndexCache():
                if len(j.getNodeIndexCache()['subjobs:status']) > 0:
                    for sj_stat in j.getNodeIndexCache()['subjobs:status']:
                        if sj_stat == status:
                            tot_active += 1
                else:
                    if j.getNodeIndexCache()['status'] == status:
                        tot_active += 1

            else:
                #logger.warning("WARNING: (status check) No index cache for job object %d" % jid)
                if j.subjobs:
                    for sj in j.subjobs:
                        if sj.status == status:
                            tot_active += 1
                else:
                    if j.status == status:
                        tot_active += 1

        return tot_active

    def n_all(self):
        total = 0
        for jid in self.active_job_ids:

            try:
                job = getJobByID(jid)
            except Exception as err:
                logger.debug("n_all Err: %s" % str(err))
                task = self._getParent()._getParent()
                trf = self._getParent()
                logger.warning("Cannot find job with id %d. Maybe reset this unit with: tasks(%d).transforms[%d].resetUnit(%d)" %
                               (jid, task.id, trf.getID(), self.getID()))
                continue

            j = stripProxy(job)

            # try to preserve lazy loading
            if hasattr(j, 'getNodeIndexCache') and j.getNodeIndexCache() and 'subjobs:status' in j.getNodeIndexCache():
                if len(j.getNodeIndexCache()['subjobs:status']) != 0:
                    total += len(j.getNodeIndexCache()['subjobs:status'])
                else:
                    total += 1
            else:
                #logger.warning("WARNING: (status check) No index cache for job object %d" % jid)
                if j.subjobs:
                    total = len(j.subjobs)
                else:
                    total = 1

        return total

    def overview(self):
        """Print an overview of this unit"""
        o = "    Unit %d: %s        " % (self.getID(), self.name)

        for s in ["submitted", "running", "completed", "failed", "unknown"]:
            o += markup("%i   " % self.n_status(s), overview_colours[s])

        print(o)

    def copyOutput(self):
        """Copy any output to the given dataset"""
        logger.error(
            "No default implementation for Copy Output - contact plugin developers")
        return False



#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\#
import datetime
from Ganga.Core import GangaException
from Ganga.GPIDev.Schema import Schema, Version, SimpleItem
from Ganga.GPIDev.Base import GangaObject
from Ganga.GPIDev.Base.Proxy import isType, stripProxy, addProxy
from GangaDirac.Lib.Backends.DiracUtils import get_result
from Ganga.Utility.logging import getLogger
logger = getLogger()
#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\#


class BKQuery(GangaObject):

    '''Class for handling LHCb bookkeeping queries.

    Currently 4 types of queries are supported: Path, RunsByDate, Run and
    Production.  These correspond to the Dirac API methods
    DiracLHCb.bkQuery<type> (see Dirac docs for details).  


    Path formats are as follows:

    type = "Path":
    /<ConfigurationName>/<Configuration Version>/\
<Sim or Data Taking Condition>/<Processing Pass>/<Event Type>/<File Type>

    type = "RunsByDate":
     /<ConfigurationName>/<Configuration Version>/<Processing Pass>/\
<Event Type>/<File Type> 

    type = "Run":
    /<Run Number>/<Processing Pass>/<Event Type>/<File Type>
    - OR -
    /<Run Number 1>-<Run Number 2>/<Processing Pass>/<Event Type>/<File Type>

    type = "Production":
    /<ProductionID>/<Processing Pass>/<Event Type>/<File Type>

    Example Usage:

    bkq = BKQuery (
    dqflag = "All" ,
    path = "/LHCb/Collision09/Beam450GeV-VeloOpen-MagDown/Real Data/\
RecoToDST-07/90000000/DST" ,
    type = "Path" 
    ) 

    bkq = BKQuery (
    startDate = "2010-05-18" ,
    selection = "Runs" ,
    endDate = "2010-05-20" ,
    dqflag = "All" ,
    path = "/LHCb/Collision10/Real Data/90000000/RAW" ,
    type = "RunsByDate" 
    ) 

    bkq = BKQuery (
    dqflag = "All" ,
    path = "111183-126823/Real Data/Reco14/Stripping20/90000000/DIMUON.DST" ,
    type = "Run" 
    ) 

    bkq = BKQuery (
    dqflag = "All" ,
    path = "/5842/Real Data/RecoToDST-07/90000000/DST" ,
    type = "Production" 
    ) 

    then (for any type) one can get the data set by doing the following:
    data = bkq.getDataset()

    This will query the bookkeeping for the up-to-date version of the data.
    N.B. BKQuery objects can be stored in your Ganga box.

    '''
    schema = {}
    docstr = 'Bookkeeping query path (type dependent)'
    schema['path'] = SimpleItem(defvalue='', doc=docstr)
    docstr = 'Start date string yyyy-mm-dd (only works for type="RunsByDate")'
    schema['startDate'] = SimpleItem(defvalue='', doc=docstr)
    docstr = 'End date string yyyy-mm-dd (only works for type="RunsByDate")'
    schema['endDate'] = SimpleItem(defvalue='', doc=docstr)
    docstr = 'Data quality flag (string or list of strings).'
    schema['dqflag'] = SimpleItem(defvalue='OK', typelist=['str', 'list'], doc=docstr)
    docstr = 'Type of query (Path, RunsByDate, Run, Production)'
    schema['type'] = SimpleItem(defvalue='Path', doc=docstr)
    docstr = 'Selection criteria: Runs, ProcessedRuns, NotProcessed (only works for type="RunsByDate")'
    schema['selection'] = SimpleItem(defvalue='', doc=docstr)
    _schema = Schema(Version(1, 2), schema)
    _category = 'query'
    _name = "BKQuery"
    _exportmethods = ['getDataset', 'getDatasetMetadata']

    def __init__(self, path=''):
        super(BKQuery, self).__init__()
        self.path = path

    def __construct__(self, args):
        if (len(args) != 1) or (not isType(args[0], str)):
            super(BKQuery, self).__construct__(args)
        else:
            self.path = args[0]

    def getDatasetMetadata(self):
        '''Gets the dataset from the bookkeeping for current path, etc.'''
        if not self.path:
            return None
        if not self.type in ['Path', 'RunsByDate', 'Run', 'Production']:
            raise GangaException('Type="%s" is not valid.' % self.type)
        if not self.type is 'RunsByDate':
            if self.startDate:
                msg = 'startDate not supported for type="%s".' % self.type
                raise GangaException(msg)
            if self.endDate:
                msg = 'endDate not supported for type="%s".' % self.type
                raise GangaException(msg)
            if self.selection:
                msg = 'selection not supported for type="%s".' % self.type
                raise GangaException(msg)
        cmd = "getDataset('%s','%s','%s','%s','%s','%s')" % (self.path, self.dqflag, self.type, self.startDate, self.endDate, self.selection)
        from Ganga.GPIDev.Lib.GangaList.GangaList import GangaList
        knownLists = [tuple, list, GangaList]
        if isType(self.dqflag, knownLists):
            cmd = "getDataset('%s',%s,'%s','%s','%s','%s')" % (self.path, self.dqflag, self.type, self.startDate, self.endDate, self.selection)

        result = get_result(cmd, 'BK query error.', 'BK query error.')
        files = []
        metadata = {}
        value = result['Value']
        if 'LFNs' in value:
            files = value['LFNs']
        if not type(files) is list:  # i.e. a dict of LFN:Metadata
            # if 'LFNs' in files: # i.e. a dict of LFN:Metadata
            metadata = files.copy()

        if metadata:
            return {'OK': True, 'Value': metadata}

        return {'OK': False, 'Value': metadata}

    def getDataset(self):
        '''Gets the dataset from the bookkeeping for current path, etc.'''
        if not self.path:
            return None
        if not self.type in ['Path', 'RunsByDate', 'Run', 'Production']:
            raise GangaException('Type="%s" is not valid.' % self.type)
        if not self.type is 'RunsByDate':
            if self.startDate:
                msg = 'startDate not supported for type="%s".' % self.type
                raise GangaException(msg)
            if self.endDate:
                msg = 'endDate not supported for type="%s".' % self.type
                raise GangaException(msg)
            if self.selection:
                msg = 'selection not supported for type="%s".' % self.type
                raise GangaException(msg)
        cmd = "getDataset('%s','%s','%s','%s','%s','%s')" % (self.path, self.dqflag, self.type, self.startDate, self.endDate, self.selection)
        from Ganga.GPIDev.Lib.GangaList.GangaList import GangaList
        knownLists = [tuple, list, GangaList]
        if isType(self.dqflag, knownLists):
            cmd = "getDataset('%s',%s,'%s','%s','%s','%s')" % (self.path, self.dqflag, self.type, self.startDate,
                                                               self.endDate, self.selection)
        result = get_result(cmd, 'BK query error.', 'BK query error.')

        logger.debug("Finished Running Command")

        files = []
        value = result['Value']
        if 'LFNs' in value:
            files = value['LFNs']
        if not type(files) is list:  # i.e. a dict of LFN:Metadata
            # if 'LFNs' in files: # i.e. a dict of LFN:Metadata
            files = files.keys()

        logger.debug("Creating DiracFile objects")

        ## Doesn't work not clear why
        from GangaDirac.Lib.Files.DiracFile import DiracFile
        #new_files = []
        #def _createDiracLFN(this_file):
        #    return DiracFile(lfn = this_file)
        #GangaObject.__createNewList(new_files, files, _createDiracLFN)

        new_files = [DiracFile(lfn=_file) for _file in files]

        #new_files = [DiracFile(lfn=_file) for _file in files]
        #for f in files:
        #    new_files.append(DiracFile(lfn=f))
            #ds.extend([DiracFile(lfn = f)])

        logger.debug("Creating Dataset")

        from GangaLHCb.Lib.LHCbDataset import LHCbDataset
        ds = LHCbDataset(new_files)

        logger.debug("Returning Dataset")

        return addProxy(ds)

#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\#


class BKQueryDict(GangaObject):

    """Class for handling LHCb bookkeeping queries using dictionaries.

    Use BKQuery if you do not know how to use BK dictionaries!

    Example Usage:

    bkqd = BKQueryDict()
    bkqd.dict['ConfigVersion'] = 'Collision09'
    bkqd.dict['ConfigName'] = 'LHCb'
    bkqd.dict['ProcessingPass'] = 'Real Data + RecoToDST-07'
    bkqd.dict['EventType'] = '90000000'
    bkqd.dict['FileType'] = 'DST'
    bkqd.dict['DataTakingConditions'] = 'Beam450GeV-VeloOpen-MagDown'
    data = bkqd.getDataset()
    """

    _bkQueryTemplate = {'SimulationConditions': 'All',
                        'DataTakingConditions': 'All',
                        'ProcessingPass': 'All',
                        'FileType': 'All',
                        'EventType': 'All',
                        'ConfigName': 'All',
                        'ConfigVersion': 'All',
                        'ProductionID':     0,
                        'StartRun':     0,
                        'EndRun':     0,
                        'DataQuality': 'All'}

    schema = {}
    docstr = 'Dirac BK query dictionary.'
    schema['dict'] = SimpleItem(defvalue=_bkQueryTemplate,  # typelist=['dict'],
                                doc=docstr)
    _schema = Schema(Version(1, 0), schema)
    _category = ''
    _name = "BKQueryDict"
    _exportmethods = ['getDataset', 'getDatasetMetadata']

    def __init__(self):
        super(BKQueryDict, self).__init__()

    def __construct__(self, args):
        if (len(args) != 1) or type(args[0]) is not dict:
            super(BKQueryDict, self).__construct__(args)
        else:
            self.dict = args[0]

    def getDatasetMetadata(self):
        '''Gets the dataset from the bookkeeping for current dict.'''
        if not self.dict:
            return None
        cmd = 'bkQueryDict(%s)' % self.dict
        result = get_result(cmd, 'BK query error.', 'BK query error.')
        files = []
        value = result['Value']
        if 'LFNs' in value:
            files = value['LFNs']
        metadata = {}
        if not type(files) is list:
            if 'LFNs' in files:  # i.e. a dict of LFN:Metadata
                metadata = files['LFNs'].copy()

        if metadata:
            return {'OK': True, 'Value': metadata}
        return {'OK': False, 'Value': metadata}

    def getDataset(self):
        '''Gets the dataset from the bookkeeping for current dict.'''
        if not self.dict:
            return None
        cmd = 'bkQueryDict(%s)' % self.dict
        result = get_result(cmd, 'BK query error.', 'BK query error.')
        files = []
        value = result['Value']
        if 'LFNs' in value:
            files = value['LFNs']
        if not type(files) is list:
            if 'LFNs' in files:  # i.e. a dict of LFN:Metadata
                files = files['LFNs'].keys()

        from GangaDirac.Lib.Files.DiracFile import DiracFile
        this_list = [DiracFile(lfn=_file) for _file in files]

        from GangaLHCb.Lib.LHCbDataset import LHCbDataset
        ds = LHCbDataset(this_list)

        return addProxy(ds)

#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\#



"""Wrapper around the publish features of stomp.py."""

import time
from Queue import Queue
import logging
logging_DEBUG = logging.DEBUG

BEAT_TIME = 0.3 # Seconds between publisher thread heart beats.
IDLE_TIMEOUT = 30 # Maximum seconds to idle before closing connection.
EXIT_TIMEOUT = 5 # Maximum seconds to clear queued messages on exit.
PUBLISHER_TIMESTAMP_HEADER = '_publisher_timestamp' # The publisher timestamp header name

try:
    import stomp
    stomp_listener = stomp.ConnectionListener
    stomp_major_version = int(stomp.__version__[0])


except (Exception, ImportError) as err:

    if not isinstance(err, ImportError):
        print("Error Importing Stomp utility!")
        print("err: %s" % err)

    class stomp(object):
        def __init__(self):
            ## DUMMY CLASS
            pass

        @staticmethod
        def Connection(self, _server_and_port={}, user='', password=''):
            ## DO NOTHING
            pass

    stomp_listener = object
    stomp_major_version = -999


class LoggerListener(stomp_listener):
    """Connection listener which logs STOMP events."""

    def __init__(self, logger):
        self._logger = logger 

    def on_connecting(self, host_and_port):
        self._logger.debug('TCP/IP connected host=%s.' % (host_and_port,))

    def on_connected(self, headers, body):
        self._log_frame('CONNECTED', headers, body)

    def on_disconnected(self):
        self._logger.debug('TCP/IP connection lost.')

    def on_message(self, headers, body):
        self._log_frame('MESSAGE', headers, body)

    def on_receipt(self, headers, body):
        self._log_frame('RECEIPT', headers, body)

    def on_error(self, headers, body):
        self._log_frame('ERROR', headers, body)

    def _log_frame(self, frame_type, headers, body):
        if self._logger.isEnabledFor(logging_DEBUG):
            self._logger.debug('STOMP %s frame received headers=%s body=%s.' % (frame_type, headers, body))

# counter to give unique thread names
_thread_id = 0

def createPublisher(T, server, port, user='', password='', logger=None,
                    idle_timeout=IDLE_TIMEOUT):
    """Create a new asynchronous publisher for sending messages to an MSG server.
    
    @param T: The thread class from which the publisher should inherit.
    @param server: The server host name.
    @param user: The user name.
    @param password: The password.
    @param logger: The logger instance.
    @param idle_timeout: Maximum seconds to idle before closing connection.
            Negative value indicates never close connection.
    
    The send method adds messages to a local queue, and the publisher thread sends
    them to the MSG server to emulate fast asynchronous sending of messages.
    
    Usage example with regular thread::
        from threading import Thread
        p = createPublisher(Thread, 'gridmsg001.cern.ch', 6163)
        p.start()
        p.addExitHandler()
        p.send('/topic/ganga.dashboard.test', 'Hello World!')
    
    Usage example with managed thread::
        #All GangaThread are registered with GangaThreadPool
        from Ganga.Core.GangaThread import GangaThread
        p = createPublisher(GangaThread, 'gridmsg001.cern.ch', 6163)
        p.start()
        p.send('/topic/ganga.dashboard.test', 'Hello World!')
        #GangaThreadPool calls p.stop() during shutdown
    """
    
    # increment thread id
    global _thread_id
    _thread_id += 1

    class AsyncStompPublisher(T):
        """Asynchronous asynchronous publisher for sending messages to an MSG server."""

        def __init__(self):
            T.__init__(self, name=('AsyncStompPublisher_%s_%s:%s' % (_thread_id, server, port)))
            self.setDaemon(True)
            # indicates that the publisher thread should exit
            self.__should_stop = False
            # indicates that publisher is currently sending
            self.__sending = False
            # connection
            self._cx = None
            # connection parameters
            self._cx_params = ([(server, port)], user, password)
            self._cx_hostname_ports = [(server, port)]
            self._cx_username = user
            self._cx_password = password
            # logger
            self._logger = logger
            # indicates how long (seconds) the connection can idle
            self.idle_timeout = idle_timeout
            # indicates initial back-off time, multiplier and maximum time
            # in case of connect/send error
            # e.g retry after 5, 10, 20, 40, 60, 60, ... seconds
            self.backoff_initial = 5
            self.backoff_multiplier = 2
            self.backoff_max = 60
            # queue to hold (message, headers, keyword_headers) tuples
            self._message_queue = Queue()
            self.__finalized = False

        def send(self, destination=None, message='', headers=None, **keyword_headers):
            """Add message to local queue for sending to MSG server.
            
            @param destination: An MSG topic or queue, e.g. /topic/dashboard.test.
            @param message: A string or dictionary of key-value pairs.
            @param headers: A dictionary of headers.
            @param keyword_headers: A dictionary of headers defined by keyword.
            
            If destination is not None, then it is added to keyword_headers.
            If not already present, then _publisher_timestamp (time since the Epoch
            (00:00:00 UTC, January 1, 1970) in seconds) is added to keyword_headers.
            Finally headers and keyword_headers are passed to stomp.py, which merges them.
            N.B. keyword_headers take precedence over headers.
            """
            if self.should_stop():
                self._log(logging_DEBUG, 'Request to queue message during or after thread shutdown denied.')
                return
            if headers is None:
                headers = {}
            if destination is not None:
                keyword_headers['destination'] = destination
            if not PUBLISHER_TIMESTAMP_HEADER in keyword_headers:
                keyword_headers[PUBLISHER_TIMESTAMP_HEADER] = time.time()
            m = (message, headers, keyword_headers)
            self._log(logging_DEBUG, 'Queuing message. body=%r headers=%r keyword_headers=%r.', *m)
            self._message_queue.put(m)
            self._log(logging_DEBUG, 'Message queued. %s queued message(s).', self._message_queue.qsize())

        def _send(self, (message, headers, keyword_headers)):
            """Send given message to MSG server."""
            if self._cx is None:
                self._log(logging_DEBUG, 'NOT Sending message:\n%s' % message)
                return
            self._log(logging_DEBUG, 'Sending message. body=%r headers=%r keyword_headers=%r.', message, headers, keyword_headers)
            global stomp_major_version
            if stomp_major_version > 2:
                import copy
                keyword_headers2 = copy.deepcopy(keyword_headers)
                my_destination = keyword_headers['destination']
                del keyword_headers2['destination']
                self._cx.send(my_destination, message, None, headers, **keyword_headers2)
            else:
                self._cx.send(message, headers, **keyword_headers)
            self._log(logging_DEBUG, 'Sent message.')

        def _connect(self):
            """Connects to MSG server if not already connected."""
            cx = self._cx
            if cx is None or not cx.is_connected():
                self._log(logging_DEBUG, 'Connecting.')
                # create connection
                global stomp_major_version
                if stomp_major_version > 2:
                    cx = stomp.Connection(self._cx_hostname_ports)
                else:
                    cx = stomp.Connection(*self._cx_params)

                if cx is None:
                    self._cx = None
                    return
                # add logger listener to connection
                if self._logger is not None:
                    cx.set_listener('logger', LoggerListener(self._logger))
                cx.start()
                if stomp_major_version > 2:
                    cx.connect(username=self._cx_username, passcode=self._cx_password)
                else:
                    cx.connect()
                self._cx = cx
                self._log(logging_DEBUG, 'Connected.')

        def _disconnect(self):
            """Disconnects (quietly) from MSG server if not already disconnected."""
            cx = self._cx
            if cx is not None:
                self._log(logging_DEBUG, 'Disconnecting')
                self._cx = None
                if cx.is_connected():
                    try:
                        cx.disconnect()
                    except Exception:
                        self._log(logging_DEBUG, 'Exception on disconnect.', exc_info=True)
                self._log(logging_DEBUG, 'Disconnected')

        def run(self):
            """Send messages, connecting as necessary and disconnecting after idle_timeout
            seconds.
            """
            # indicates time in seconds since last connect/send attempt
            idle_time = 0
            # indicates time in seconds before next connect/send attempt
            backoff_time = 0
            # run unless should_stop and (queue empty or backing off)
            while not (self.should_stop() and (self._message_queue.empty() or backoff_time > 0)):
                # if idle_time exceeds backoff_time then attempt to connect/send
                if idle_time >= backoff_time:
                    # send unless queue empty
                    while not self._message_queue.empty():
                        try:
                            self._log(logging_DEBUG, 'Before attempt to connect/send. %s queued message(s).', self._message_queue.qsize())
                            self.__sending = True
                            idle_time = 0
                            m = None
                            try:
                                self._connect()
                                m = self._message_queue.get()
                                self._send(m)
                                # reset backoff_time
                                backoff_time = 0
                            except Exception:
                                self._log(logging_DEBUG, 'Exception on connect/send.', exc_info=True)
                                # increment backoff_time
                                if backoff_time == 0:
                                    backoff_time = self.backoff_initial
                                else:
                                    backoff_time = min(backoff_time * self.backoff_multiplier, self.backoff_max)
                                self._log(logging_DEBUG, 'Back-off time set to %s seconds.', backoff_time)
                                # re-queue message
                                if m is not None:
                                    self._log(logging_DEBUG, 'Re-queuing failed message. body=%r headers=%r keyword_headers=%r.', *m)
                                    self._message_queue.put(m)
                                break # break out to wait for BEAT_TIME
                        finally:
                            self.__sending = False
                            self._log(logging_DEBUG, 'After attempt to connect/send. %s queued message(s).', self._message_queue.qsize())
                # heart beat pause
                time.sleep(BEAT_TIME)
                # fix for bug #62543 https://savannah.cern.ch/bugs/?62543
                # exit directly if python interpreter has torn down globals
                if stomp is None:
                    return
                # increment idle_time
                idle_time += BEAT_TIME
                # disconnect (quietly) if idle_timeout exceeded
                if idle_time >= self.idle_timeout > -1:
                    self._disconnect()
            # disconnect (quietly)
            self._disconnect()
            self._log(logging_DEBUG, 'Exit publisher. Local message queue size is %s.', self._message_queue.qsize())
            self._log(logging_DEBUG, "Stopped: %s" % self.getName())

        def should_stop(self):
            """Indicates whether stop() has been called."""
            return self.__should_stop

        def stop(self):
            """Tells the publisher thread to stop gracefully.
            
            Typically called on a managed thread such as GangaThread.
            """
            if not self.__should_stop:
                self._log(logging_DEBUG, "Stopping: %s.", self.getName())
                self.__should_stop = True

        def addExitHandler(self, timeout=EXIT_TIMEOUT):
            """Adds an exit handler that allows the publisher up to timeout seconds to send
            queued messages.
            
            @param timeout: Maximum seconds to clear message queue on exit.
                    Negative value indicates clear queue without timeout.
            
            Typically used on unmanaged threads, i.e. regular Thread not GangaThread.
            """
            # register atexit finalize method
            import atexit
            atexit.register(self._finalize, timeout)

        def _finalize(self, timeout):
            """Allow the publisher thread up to timeout seconds to send queued messages.
            
            @param timeout: Maximum seconds to clear message queue on exit.
                    Negative value indicates clear queue without timeout.
            """
            if self.__finalized is True:
                return
            self._log(logging_DEBUG, 'Finalizing.')
            finalize_time = 0
            while not self._message_queue.empty() or self.__sending:
                if finalize_time >= timeout > -1:
                    break
                time.sleep(BEAT_TIME)
                finalize_time += BEAT_TIME
            self._disconnect()
            self._log(logging_DEBUG, 'Finalized after %s second(s). Local message queue size is %s.', finalize_time, self._message_queue.qsize())
            self.__finalized = True

        def _log(self, level, msg, *args, **kwargs):
            """Log message if logger is defined."""
            if self._logger is not None:
                self._logger.log(*((level, msg,) + args), **kwargs)

    return AsyncStompPublisher()


# for testing purposes
if __name__ == '__main__':
    from Ganga.Utility.logging import getLogger
    l = getLogger()
    from threading import Thread
    p = createPublisher(Thread, 'gridmsg001.cern.ch', 6163, logger=l, idle_timeout=5)
    p.start()
    p.addExitHandler(5)
    p.send('/topic/ganga.dashboard.test', 'Hello World 1!')
    p.send('/topic/ganga.dashboard.test', 'Hello World 2!')
    time.sleep(10)
    p.send('/topic/ganga.dashboard.test', 'Hello World 3!', {'foo1': 'bar1'}, foo2='bar2')
    p.send('/topic/ganga.dashboard.test', 'Hello World 4!', foo2='bar2')
    p.send('/topic/ganga.dashboard.test', repr({'msg': 'Hello World 5!'}))



from Ganga.Utility.Config import makeConfig

# test configuration properties
test_config = makeConfig('TestConfig','testing stuff')
test_config.addOption('None_OPT', None, '')
test_config.addOption('Int_OPT', 1, '')
test_config.addOption('List_OPT', [1,2,3], '')
test_config.addOption('String_OPT' ,'dupa', '')
# there is also an Undefine_OPT which will be used in the test case



from __future__ import absolute_import
from __future__ import print_function
from Ganga.GPIDev.Base import GangaObject
from Ganga.GPIDev.Schema import Schema, Version, SimpleItem, ComponentItem, FileItem, GangaFileItem
from Ganga.Utility.logging import getLogger
from Ganga.Utility.ColourText import status_colours, overview_colours, ANSIMarkup
markup = ANSIMarkup()
from Ganga.Core.exceptions import ApplicationConfigurationError
from Ganga.GPIDev.Lib.Job import MetadataDict
from Ganga.Utility.Config import getConfig
from Ganga.GPIDev.Base.Proxy import stripProxy, isType, getName
from .IUnit import IUnit
import time
import os
from Ganga.GPIDev.Lib.Tasks.ITask import addInfoString
from Ganga.GPIDev.Lib.Tasks.common import getJobByID
from Ganga.GPIDev.Adapters.IGangaFile import IGangaFile
from Ganga.GPIDev.Lib.File.File import File

logger = getLogger()

class ITransform(GangaObject):
    _schema = Schema(Version(1, 0), {
        'status': SimpleItem(defvalue='new', protected=1, copyable=0, doc='Status - running, pause or completed', typelist=[str]),
        'name': SimpleItem(defvalue='Simple Transform', doc='Name of the transform (cosmetic)', typelist=[str]),
        'application': ComponentItem('applications', defvalue=None, optional=1, load_default=False, doc='Application of the Transform.'),
        'inputsandbox': FileItem(defvalue=[], sequence=1, doc="list of File objects shipped to the worker node "),
        'outputsandbox': SimpleItem(defvalue=[], typelist=[str], sequence=1, doc="list of filenames or patterns shipped from the worker node"),
        'backend': ComponentItem('backends', defvalue=None, optional=1, load_default=False, doc='Backend of the Transform.'),
        'splitter': ComponentItem('splitters', defvalue=None, optional=1, load_default=False, doc='Splitter used on each unit of the Transform.'),
        'postprocessors': ComponentItem('postprocessor', defvalue=None, doc='list of postprocessors to run after job has finished'),
        'merger': ComponentItem('mergers', defvalue=None, hidden=1, copyable=0, load_default=0, optional=1, doc='Merger to be done over all units when complete.'),
        'unit_merger': ComponentItem('mergers', defvalue=None, load_default=0, optional=1, doc='Merger to be copied and run on each unit separately.'),
        'copy_output': ComponentItem('datasets', defvalue=None, load_default=0, optional=1, doc='The dataset to copy all units output to, e.g. Grid dataset -> Local Dataset'),
        'unit_copy_output': ComponentItem('datasets', defvalue=None, load_default=0, optional=1, doc='The dataset to copy each individual unit output to, e.g. Grid dataset -> Local Dataset'),
        'run_limit': SimpleItem(defvalue=8, doc='Number of times a partition is tried to be processed.', protected=1, typelist=[int]),
        'minor_run_limit': SimpleItem(defvalue=3, doc='Number of times a unit can be resubmitted', protected=1, typelist=[int]),
        'major_run_limit': SimpleItem(defvalue=3, doc='Number of times a junit can be rebrokered', protected=1, typelist=[int]),
        'units': ComponentItem('units', defvalue=[], sequence=1, copyable=1, doc='list of units'),
        'inputdata': ComponentItem('datasets', defvalue=[], sequence=1, protected=1, optional=1, load_default=False, doc='Input datasets to run over'),
        'outputdata': ComponentItem('datasets', defvalue=None, optional=1, load_default=False, doc='Output dataset template'),
        'inputfiles': GangaFileItem(defvalue=[], sequence=1, doc="list of file objects that will act as input files for a job"),
        'outputfiles' : GangaFileItem(defvalue=[], sequence=1, doc="list of OutputFile objects to be copied to all jobs"),
        'metadata': ComponentItem('metadata', defvalue=MetadataDict(), doc='the metadata', protected=1),
        'rebroker_on_job_fail': SimpleItem(defvalue=True, doc='Rebroker if too many minor resubs'),
        'abort_loop_on_submit': SimpleItem(defvalue=True, doc='Break out of the Task Loop after submissions'),
        'required_trfs': SimpleItem(defvalue=[], typelist=[int], sequence=1, doc="IDs of transforms that must complete before this unit will start. NOTE DOESN'T COPY OUTPUT DATA TO INPUT DATA. Use TaskChainInput Dataset for that."),
        'chain_delay': SimpleItem(defvalue=0, doc='Minutes delay between a required/chained unit completing and starting this one', protected=0, typelist=[int]),
        'submit_with_threads': SimpleItem(defvalue=False, doc='Use Ganga Threads for submission'),
        'max_active_threads': SimpleItem(defvalue=10, doc='Maximum number of Ganga Threads to use. Note that the number of simultaneous threads is controlled by the queue system (default is 5)'),
        'info' : SimpleItem(defvalue=[],typelist=[str],protected=1,sequence=1,doc="Info showing status transitions and unit info"),
        'id': SimpleItem(defvalue=-1, protected=1, doc='ID of the Transform', typelist=[int]),
        #'force_single_unit' : SimpleItem(defvalue=False, doc='Force all input data into one Unit'),
    })

    _category = 'transforms'
    _name = 'ITransform'
    _exportmethods = ['addInputData', 'resetUnit', 'setRunLimit', 'getJobs', 'setMinorRunLimit',
                      'setMajorRunLimit', 'getID', 'overview', 'resetUnitsByStatus', 'removeUnusedJobs',
                      'showInfo', 'showUnitInfo', 'pause', 'n_all', 'n_status' ]
    _hidden = 0

    def showInfo(self):
        """Print out the info in a nice way"""
        print("\n".join( self.info ))

    def showUnitInfo(self, uid):
        """Print out the given unit info in a nice way"""
        self.units[uid].showInfo()

    def getJobs(self):
        """Return a list of the currently active job ids"""
        joblist = []
        for u in self.units:
            joblist += u.active_job_ids
        return joblist

    def setMinorRunLimit(self, newRL):
        """Set the number of times a job will be resubmitted before a major resubmit is attempted"""
        self.minor_run_limit = newRL

    def setMajorRunLimit(self, newRL):
        """Set the number of times a job will be rebrokered before the transform is paused"""
        self.major_run_limit = newRL

    def setRunLimit(self, newRL):
        """Set the total (minor+major) number of times a job should be resubmitted before the transform is paused"""
        self.run_limit = newRL

    def overview(self, status=''):
        """Show the status of the units in this transform"""
        for unit in self.units:
            # display colour given state
            o = ""
            o += ("%d:  " % self.units.index(unit)) + unit.name

            # is unit active?
            if unit.active:
                o += " " * (40 - len(o) + 3) + "*"
            else:
                o += " " * (40 - len(o) + 3) + "-"

            # sub job status
            o += "\t %i" % unit.n_status("submitted")
            o += "\t %i" % unit.n_status("running")
            o += "\t %i" % unit.n_status("completed")
            o += "\t %i" % unit.n_status("failed")
            o += "\t %i" % unit.minor_resub_count
            o += "\t %i" % unit.major_resub_count

            # change colour on state
            if unit.status == 'completed':
                o = markup(o, overview_colours["completed"])
            elif not unit.active:
                o = markup(o, overview_colours["bad"])
            elif unit.status == "recreating":
                o = markup(o, overview_colours["attempted"])
            elif len(unit.active_job_ids) == 0:
                o = markup(o, overview_colours["hold"])
            else:
                o = markup(o, overview_colours["running"])

            print(o)


# Special methods:
    def __init__(self):
        super(ITransform, self).__init__()
        self.initialize()

    def _readonly(self):
        """A transform is read-only if the status is not new."""
        if self.status == "new":
            return 0
        return 1

    def initialize(self):
        from Ganga.Lib.Localhost.Localhost import Localhost
        self.backend = Localhost()
        self.updateStatus("new")

    def check(self):
        """Check this transform has valid data, etc. and has the correct units"""

        # ignore anything but new transforms
        if self.status != "new":
            return

        # first, validate the transform
        if not self.validate():
            raise ApplicationConfigurationError(
                None, "Validate failed for Transform %s" % self.name)

        self.updateStatus("running")

    def startup(self):
        """This function is used to set the status after restarting Ganga"""
        pass

# Public methods
    def resetUnit(self, uid):
        """Reset the given unit"""
        addInfoString( self, "Reseting Unit %i" % ( uid ) )

        for u in self.units:
            if u.getID() == uid:
                u.reset()
                break

        # find any chained units and mark for recreation
        for trf in self._getParent().transforms:
            for u2 in trf.units:
                for req in u2.req_units:
                    if req == "%d:%d" % (self.getID(), u.getID()) or req == "%d:ALL" % (self.getID()):
                        trf.resetUnit(u2.getID())

        self.updateStatus("running")

    def getID(self):
        """Return the index of this trf in the parent task"""

        # if the id isn't already set, use the index from the parent Task
        if self.id < 0:
            task = self._getParent()
            if not task:
                raise ApplicationConfigurationError(
                    None, "This transform has not been associated with a task and so there is no ID available")
            self.id = task.transforms.index(self)
        
        return self.id

    def run(self, check=True):
        """Sets this transform to running status"""
        if self.status == "new" and check:
            self.check()
        if self.status != "completed":
            self.updateStatus("running")
            task = self._getParent()
            if task:
                task.updateStatus()
        else:
            logger.warning("Transform is already completed!")

    def update(self):
        """Called by the parent task to check for status updates, submit jobs, etc."""
        if self.status == "pause" or self.status == "new":
            return 0

        # check for complete required units
        task = self._getParent()
        for trf_id in self.required_trfs:
            if task.transforms[trf_id].status != "completed":
                return 0

        # set the start time if not already set
        if len(self.required_trfs) > 0 and self.units[0].start_time == 0:
            for unit in self.units:
                unit.start_time = time.time() + self.chain_delay * 60 - 1

        # report the info for this transform
        unit_status = { "new":0, "hold":0, "running":0, "completed":0, "bad":0, "recreating":0 }
        for unit in self.units:
            unit_status[unit.status] += 1
         
        info_str = "Unit overview: %i units, %i new, %i hold, %i running, %i completed, %i bad. to_sub %i" % (len(self.units), unit_status["new"], unit_status["hold"],
                                                                                                              unit_status["running"], unit_status["completed"],
                                                                                                              unit_status["bad"], self._getParent().n_tosub())
      
        addInfoString(self, info_str)
                
        # ask the unit splitter if we should create any more units given the
        # current data
        self.createUnits()

        # loop over units and update them ((re)submits will be called here)
        old_status = self.status
        unit_status_list = []

        # find submissions first
        unit_update_list = []
        for unit in self.units:

            if not unit.checkForSubmission() and not unit.checkForResubmission():
                unit_update_list.append(unit)
                continue

            if unit.update() and self.abort_loop_on_submit:
                logger.info("Unit %d of transform %d, Task %d has aborted the loop" % (
                    unit.getID(), self.getID(), task.id))
                return 1

            unit_status_list.append(unit.status)

        # now check for download
        for unit in unit_update_list:
            if unit.update() and self.abort_loop_on_submit:
                logger.info("Unit %d of transform %d, Task %d has aborted the loop" % (
                    unit.getID(), self.getID(), task.id))
                return 1

            unit_status_list.append(unit.status)

        from Ganga.GPIDev.Lib.Tasks.TaskChainInput import TaskChainInput
        # check for any TaskChainInput completions
        for ds in self.inputdata:
            if isType(ds, TaskChainInput) and ds.input_trf_id != -1:
                if task.transforms[ds.input_trf_id].status != "completed":
                    return 0

        # update status and check
        for state in ['running', 'hold', 'bad', 'completed']:
            if state in unit_status_list:
                if state == 'hold':
                    state = "running"
                if state != self.status:
                    self.updateStatus(state)
                break

    def createUnits(self):
        """Create new units if required given the inputdata"""

        from Ganga.GPIDev.Lib.Tasks.TaskChainInput import TaskChainInput
        # check for chaining
        for ds in self.inputdata:
            if isType(ds, TaskChainInput) and ds.input_trf_id != -1:

                # check for single unit
                if ds.single_unit:

                    # is there a unit already linked?
                    done = False
                    rec_unit = None
                    for out_unit in self.units:
                        if '%d:ALL' % (ds.input_trf_id) in out_unit.req_units:
                            done = True
                            # check if the unit is being recreated
                            if out_unit.status == "recreating":
                                rec_unit = out_unit
                            break

                    if not done or rec_unit:
                        new_unit = self.createChainUnit(
                            self._getParent().transforms[ds.input_trf_id].units, ds.use_copy_output)
                        if new_unit:
                            self.addChainUnitToTRF(
                                new_unit, ds, -1, prev_unit=rec_unit)

                else:

                    # loop over units in parent trf and create units as
                    # required
                    for in_unit in self._getParent().transforms[ds.input_trf_id].units:

                        # is there a unit already linked?
                        done = False
                        rec_unit = None
                        for out_unit in self.units:
                            if '%d:%d' % (ds.input_trf_id, in_unit.getID()) in out_unit.req_units:
                                done = True
                                # check if the unit is being recreated
                                if out_unit.status == "recreating":
                                    rec_unit = out_unit
                                break

                        if not done or rec_unit:
                            new_unit = self.createChainUnit(
                                [in_unit], ds.use_copy_output)
                            if new_unit:
                                self.addChainUnitToTRF(
                                    new_unit, ds, in_unit.getID(), prev_unit=rec_unit)

    def createChainUnit(self, parent_units, use_copy_output=True):
        """Create a chained unit given the parent outputdata"""
        return IUnit()

    def addChainUnitToTRF(self, unit, inDS, unit_id=-1, prev_unit=None):
        """Add a chained unit to this TRF. Override for more control"""
        if unit_id == -1:
            unit.req_units.append('%d:ALL' % (inDS.input_trf_id))
            unit.name = "Parent: TRF %d, All Units" % (inDS.input_trf_id)
        else:
            unit.req_units.append('%d:%d' % (inDS.input_trf_id, unit_id))
            unit.name = "Parent: TRF %d, Unit %d" % (
                inDS.input_trf_id, unit_id)

        self.addUnitToTRF(unit, prev_unit)

    def addInputData(self, inDS):
        """Add the given input dataset to the list"""
        self.inputdata.append(inDS)

    def pause(self):
        """Pause the task - the background thread will not submit new jobs from this task"""
        if self.status != "completed":
            self.updateStatus("pause")
            #self.status = "pause"
            task = self._getParent()
            if task:
                task.updateStatus()
        else:
            logger.debug("Transform is already completed!")

    def setRunlimit(self, newRL):
        """Set the number of times a job should be resubmitted before the transform is paused"""
        self.run_limit = newRL
        logger.debug("Runlimit set to %i", newRL)

# Methods that can/should be overridden by derived classes
    def validate(self):
        """Override this to validate that the transform is OK"""

        from Ganga.GPIDev.Lib.Tasks.TaskLocalCopy import TaskLocalCopy
        # make sure a path has been selected for any local downloads
        if self.unit_copy_output is not None and isType(self.unit_copy_output, TaskLocalCopy):
            if self.unit_copy_output.local_location == '':
                logger.error("No path selected for Local Output Copy")
                return False

        if self.copy_output is not None and isType(self.copy_output, TaskLocalCopy):
            if self.copy_output.local_location == '':
                logger.error("No path selected for Local Output Copy")
                return False

        # this is a generic trf so assume the application and splitter will do
        # all the work
        return True

    def addUnitToTRF(self, unit, prev_unit=None):
        """Add a unit to this Transform given the input and output data"""
        if not unit:
            raise ApplicationConfigurationError(None, "addUnitTOTRF failed for Transform %d (%s): No unit specified" % (self.getID(), self.name))

        addInfoString( self, "Adding Unit to TRF...")
        unit.updateStatus("hold")
        unit.active = True
        if prev_unit:
            unit.prev_job_ids += prev_unit.prev_job_ids
            self.units[prev_unit.getID()] = unit
        else:
            self.units.append(unit)
            stripProxy(unit).id = len(self.units) - 1

# Information methods
    def fqn(self):
        task = self._getParent()
        if task:
            return "Task %i Transform %i" % (task.id, task.transforms.index(self))
        else:
            return "Unassigned Transform '%s'" % (self.name)

    def n_active(self):
        return sum([u.n_active() for u in self.units])

    def n_all(self):
        return sum([u.n_all() for u in self.units])

    def n_status(self, status):
        return sum([u.n_status(status) for u in self.units])

    def info(self):
        logger.info(markup("%s '%s'" % (getName(self), self.name), status_colours[self.status]))
        logger.info("* backend: %s" % getName(self.backend))
        logger.info("Application:")
        self.application.printTree()

    def updateStatus(self, status):
        """Update the transform status"""
        self.status = status

    def createUnitCopyOutputDS(self, unit_id):
        """Create a the Copy Output dataset to use with this unit. Overload to handle more than the basics"""

        from Ganga.GPIDev.Lib.Tasks.TaskLocalCopy import TaskLocalCopy
        if isType(self.unit_copy_output, TaskLocalCopy):
            logger.warning("Default implementation of createUnitCopyOutputDS can't handle datasets of type '%s'" % getName(self.unit_copy_output))
            return

        # create copies of the Copy Output DS and add Unit name to path
        self.units[unit_id].copy_output = self.unit_copy_output.clone()
        self.units[unit_id].copy_output.local_location = os.path.join(
            self.unit_copy_output.local_location, self.units[unit_id].name.replace(":", "_").replace(" ", "").replace(",", "_"))

    def __setattr__(self, attr, value):

        if attr == 'outputfiles':

            if value != []:
                if self.outputdata is not None:
                    logger.error(
                        'ITransform.outputdata is set, you can\'t set ITransform.outputfiles')
                    return
                elif self.outputsandbox != []:
                    logger.error(
                        'ITransform.outputsandbox is set, you can\'t set ITransform.outputfiles')
                    return

            # reduce duplicate values here, leave only duplicates for LCG,
            # where we can have replicas
            uniqueValuesDict = []
            uniqueValues = []

            for val in value:
                key = '%s%s' % (getName(val), val.namePattern)
                if key not in uniqueValuesDict:
                    uniqueValuesDict.append(key)
                    uniqueValues.append(val)
                elif getName(val) == 'LCGSEFile':
                    uniqueValues.append(val)

            super(ITransform, self).__setattr__(attr, uniqueValues)

        elif attr == 'inputfiles':

            if value != []:
                if self.inputsandbox != []:
                    logger.error(
                        'ITransform.inputsandbox is set, you can\'t set ITransform.inputfiles')
                    return

            super(ITransform, self).__setattr__(attr, value)

        elif attr == 'outputsandbox':

            if value != []:

                if getConfig('Output')['ForbidLegacyOutput']:
                    logger.error(
                        'Use of ITransform.outputsandbox is forbidden, please use ITransform.outputfiles')
                    return

                if self.outputfiles != []:
                    logger.error(
                        'ITransform.outputfiles is set, you can\'t set ITransform.outputsandbox')
                    return

            super(ITransform, self).__setattr__(attr, value)

        elif attr == 'inputsandbox':

            if value != []:

                if getConfig('Output')['ForbidLegacyInput']:
                    logger.error(
                        'Use of ITransform.inputsandbox is forbidden, please use ITransform.inputfiles')
                    return

                if self.inputfiles != []:
                    logger.error(
                        'ITransform.inputfiles is set, you can\'t set ITransform.inputsandbox')
                    return

            super(ITransform, self).__setattr__(attr, value)

        elif attr == 'outputdata':

            if value is not None:

                if getConfig('Output')['ForbidLegacyOutput']:
                    logger.error(
                        'Use of ITransform.outputdata is forbidden, please use ITransform.outputfiles')
                    return

                if self.outputfiles != []:
                    logger.error(
                        'ITransform.outputfiles is set, you can\'t set ITransform.outputdata')
                    return
            super(ITransform, self).__setattr__(attr, value)

        else:
            super(ITransform, self).__setattr__(attr, value)

    def resetUnitsByStatus(self, status='bad'):
        """Reset all units of a given status"""
        for unit in self.units:
            if unit.status == status:
                logger.info("Resetting Unit %d, Transform %d..." %
                            (unit.getID(), self.getID()))
                self.resetUnit(unit.getID())

    def checkUnitsAreCompleted(self, parent_units):
        """Check the given parent units are complete"""
        for parent in parent_units:
            if len(parent.active_job_ids) == 0 or parent.status != "completed":
                return False

        return True

    def getChainInclExclMasks(self, parent_units):
        """return the include/exclude masks from the TaskChainInput"""
        incl_pat_list = []
        excl_pat_list = []
        from Ganga.GPIDev.Lib.Tasks.TaskChainInput import TaskChainInput
        for parent in parent_units:
            for inds in self.inputdata:
                if isType(inds, TaskChainInput) and inds.input_trf_id == parent._getParent().getID():
                    incl_pat_list += inds.include_file_mask
                    excl_pat_list += inds.exclude_file_mask

        return incl_pat_list, excl_pat_list

    def getParentUnitJobs(self, parent_units, include_subjobs=True):
        """Return the list of parent jobs"""
        job_list = []
        for parent in parent_units:
            job = getJobByID(parent.active_job_ids[0])
            if job.subjobs:
                job_list += job.subjobs
            else:
                job_list += [job]

        return job_list

    def removeUnusedJobs(self):
        """Remove all jobs that aren't being used, e.g. failed jobs"""
        for unit in self.units:
            for jid in unit.prev_job_ids:
                try:
                    logger.warning("Removing job '%d'..." % jid)
                    job = getJobByID(jid)
                    job.remove()
                except Exception as err:
                    logger.debug("removeUnused: %s" % str(err))
                    logger.error("Problem removing job '%d'" % jid)



#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\#
'''Parent for all Gaudi and GaudiPython applications in LHCb.'''

import os
import tempfile
import gzip
import shutil
from Ganga.GPIDev.Base.Proxy import stripProxy
from Ganga.GPIDev.Schema import SimpleItem, Schema, Version
from Ganga.GPIDev.Adapters.IPrepareApp import IPrepareApp
import Ganga.Utility.logging
from Ganga.Utility.files import expandfilename, fullpath
from GaudiUtils import get_user_platform, fillPackedSandbox, get_user_dlls
from Ganga.GPIDev.Lib.File import File
from Ganga.Core import ApplicationConfigurationError
import Ganga.Utility.Config
from Ganga.Utility.execute import execute
from Ganga.GPIDev.Lib.File import ShareDir
from Ganga.Utility.Config import getConfig
from Ganga.GPIDev.Base.Proxy import getName
import copy
logger = Ganga.Utility.logging.getLogger()


class GaudiBase(IPrepareApp):

    '''Parent for all Gaudi and GaudiPython applications, should not be used
    directly.'''

    schema = {}
    docstr = 'The version of the application (like "v19r2")'
    schema['version'] = SimpleItem(preparable=1, defvalue=None,
                                   typelist=['str', 'type(None)'], doc=docstr)
    docstr = 'The platform the application is configured for (e.g. ' \
             '"slc4_ia32_gcc34")'
    schema['platform'] = SimpleItem(preparable=1, defvalue=None,
                                    typelist=['str', 'type(None)'], doc=docstr)
    docstr = 'The user path to be used. After assigning this'  \
             ' you can do j.application.getpack(\'Phys DaVinci v19r2\') to'  \
             ' check out into the new location. This variable is used to '  \
             'identify private user DLLs by parsing the output of "cmt '  \
             'show projects".'
    schema['user_release_area'] = SimpleItem(preparable=1, defvalue=None,
                                             typelist=['str', 'type(None)'],
                                             doc=docstr)
    docstr = 'The name of the Gaudi application (e.g. "DaVinci", "Gauss"...)'
    schema['appname'] = SimpleItem(preparable=1, defvalue=None, typelist=['str', 'type(None)'],
                                   hidden=1, doc=docstr)
    docstr = 'Location of shared resources. Presence of this attribute implies'\
             'the application has been prepared.'
    schema['is_prepared'] = SimpleItem(defvalue=None,
                                       strict_sequence=0,
                                       visitable=1,
                                       copyable=1,
                                       typelist=['type(None)', 'str', ShareDir],
                                       hidden=0,
                                       protected=1,
                                       doc=docstr)
    docstr = 'The env'
    schema['env'] = SimpleItem(preparable=1, transient=1, defvalue=None,
                               hidden=1, doc=docstr, typelist=['type(None)', 'dict'])
    docstr = 'MD5 hash of the string representation of applications preparable attributes'
    schema['hash'] = SimpleItem(defvalue=None, typelist=['type(None)', 'str'], hidden=1)

    schema['newStyleApp'] = SimpleItem(defvalue=False, typelist=['bool'], doc="Is this app a 'new Style' CMake app?")

    _name = 'GaudiBase'
    _exportmethods = ['getenv', 'getpack', 'make', 'projectCMD', 'cmt']
    _schema = Schema(Version(0, 1), schema)
    _hidden = 1

    
    #def __init__(self):
    #    super(GaudiBase, self).__init__(None)

    def _get_default_version(self, gaudi_app):
        raise NotImplementedError

    def _get_default_platform(self):
        return get_user_platform(self)

    #def _init(self, set_ura):
    def _init(self):
        if self.appname is None:
            raise ApplicationConfigurationError(None, "appname is None")
        if (not self.version):
            self.version = self._get_default_version(self.appname)
        if (not self.platform):
            self.platform = self._get_default_platform()
        #if not set_ura:
        #    return
        if not self.user_release_area:
            expanded = os.path.expandvars("$User_release_area")
            if expanded == "$User_release_area":
                self.user_release_area = ""
            else:
                self.user_release_area = expanded.split(os.pathsep)[0]

        from Ganga.Utility.Shell import expand_vars
        env = expand_vars(os.environ)

        env['User_release_area'] = self.user_release_area
        env['CMTCONFIG'] = self.platform
        return env

    def getenv(self, cache_env=False):
        '''Returns a copy of the environment used to flatten the options, e.g.
        env = DaVinci().getenv(), then calls like env[\'DAVINCIROOT\'] return
        the values.

        Note: Editing this does not affect the options processing.
        '''
        if self.env is None:
            shell = None
            try:
                job = self.getJobObject()
            except Exception as err:
                logger.debug("Error: %s" % str(err))
                pass
            else:
                env_file_name = job.getDebugWorkspace().getPath() + \
                    '/gaudi-env.py.gz'
                if os.path.exists(env_file_name):
                    in_file = gzip.GzipFile(env_file_name, 'rb')
                    exec(in_file.read())
                    in_file.close()
                    shell = gaudi_env

            if shell is None:
                shell = self._getshell()
            if cache_env:
                self.env = copy.deepcopy(shell)
            return shell

        return self.env

    def _export_getenv(self):
        return self.getenv(False)

    def getpack(self, options=''):
        """Execute a getpack command. If as an example dv is an object of
        type DaVinci, the following will check the Analysis package out in
        the project area pointed to by the dv object.

        dv.getpack('Tutorial/Analysis v6r2')
        """
        # Make sure project user area is there
        project_path = expandfilename(self.user_release_area)
        if project_path:
            if not os.path.exists(project_path):
                try:
                    os.makedirs(project_path)
                except Exception as err:
                    logger.error("Can not create project user directory: " + project_path)
                    logger.debug("%s" % str(err))
                    return

        execute('getpack %s' % options,
                shell=True,
                timeout=None,
                env=self.getenv(False),
                cwd=self.user_release_area)

    def make(self, argument=''):
        """Build the code in the release area the application object points to."""
        from GangaGaudi.Lib.Application.GaudiUtils import make
        make(self, argument)

    def projectCMD(self, command):
        """Eecute a given command at the top level of a requested project."""

        execute('%s' % command,
                shell=True,
                timeout=None,
                env=self.getenv(False),
                cwd=self.user_release_area)

    def cmt(self, command):
        """Execute a cmt command in the cmt user area pointed to by the
        application. Will execute the command "cmt <command>" after the
        proper configuration. Do not include the word "cmt" yourself."""

        if self.newStyleApp is True:
            logger.warning("Cannot Use this in combination with cmake!")
            return

        execute('cmt %s' % command,
                shell=True,
                timeout=None,
                env=self.getenv(False),
                cwd=self.user_release_area)

    def unprepare(self):
        self._unregister()

    def _unregister(self):
        if self.is_prepared is not None:
            self.decrementShareCounter(self.is_prepared.name)
            self.is_prepared = None
        self.hash = None
        self.env = None

    def prepare(self, force=False):
        # Define the ShareDir
        self._register(force)
        try:
            # Try to prepare the application fully
            self._really_prepare(force)
        except Exception, err:
            # Cleanup after self on fail as self.is_prepared is used as test to
            # see if I'm prepared
            logger.error("Prepare Error:\n%s" % str(err))
            self._unregister()
            raise

    def _really_prepare(self, force=False):
        if (not self.is_prepared):
            raise ApplicationConfigurationError(
                None, "Could not establish sharedir")

        if (not self.user_release_area):
            return  # GaudiPython and Bender dont need the following.
        if (not self.appname):
            raise ApplicationConfigurationError(None, "appname is None")
        if (not self.version):
            raise ApplicationConfigurationError(None, "version is None")
        if (not self.platform):
            raise ApplicationConfigurationError(None, "platform is None")

        share_dir = os.path.join(expandfilename(getConfig('Configuration')['gangadir']),
                                 'shared',
                                 getConfig('Configuration')['user'],
                                 self.is_prepared.name)

        dlls, pys, subpys = get_user_dlls(self, self.appname, self.version,
                                          self.user_release_area, self.platform,
                                          self.getenv(True))
        InstallArea = []

        for f in dlls:
            logger.debug("UserDLLs DLLs: %s" % expandfilename(f))
            InstallArea.append(File(name=expandfilename(f), subdir='lib'))
        for f in pys:
            tmp = f.split('InstallArea')[-1]
            subdir = 'InstallArea' + tmp[:tmp.rfind('/') + 1]
            logger.debug("UserDLLs PYS: %s" % expandfilename(f))
            InstallArea.append(File(name=expandfilename(f), subdir=subdir))

        for dir, files in subpys.iteritems():
            for f in files:
                tmp = f.split('InstallArea')[-1]
                subdir = 'InstallArea' + tmp[:tmp.rfind('/') + 1]
                logger.debug("UserDLLs SUBPYS: %s" % expandfilename(f))
                InstallArea.append(File(name=expandfilename(f), subdir=subdir))

        # add the newly created shared directory into the metadata system if the app is associated with a persisted object
        # also call post_prepare for hashing
        # commented out here as inherrited from this class with extended
        # perpare

        fillPackedSandbox(InstallArea, os.path.join(share_dir, 'inputsandbox', '_input_sandbox_%s.tar' % self.is_prepared.name))

    def _register(self, force):
        if (self.is_prepared is not None) and (force is not True):
            raise Exception('%s application has already been prepared. Use prepare(force=True) to prepare again.' % (getName(self)))

        try:
            logger.info('Job %s: Preparing %s application.' % (stripProxy(self).getJobObject().getFQID('.'), getName(self)))
        except AssertionError, err:
            ## No Job associated with Object!!
            logger.info("Preparing %s application." % getName(self))
        self.is_prepared = ShareDir()

    def master_configure(self):
        '''Handles all common master_configure actions.'''
        raise NotImplementedError

    def configure(self, appmasterconfig):
        raise NotImplementedError





##############################################################################
# Ganga Project. http://cern.ch/ganga
#
# $Id: ATLASDataset.py,v 1.8 2009-04-24 08:54:23 dvanders Exp $
###############################################################################
# A simple ATLAS dataset
#
# ATLAS/ARDA

import os, re, fnmatch
import commands

from Ganga.GPIDev.Lib.Dataset import Dataset
from Ganga.GPIDev.Schema import *

from Ganga.Utility.Config  import getConfig, makeConfig, ConfigError 
from Ganga.Utility.logging import getLogger
from Ganga.Utility.Shell import Shell
from Ganga.Utility.files import expandfilename
from Ganga.Utility.GridShell import getShell

shell = Shell()
logger = getLogger()
config = getConfig('Athena')
configDQ2 = getConfig('DQ2')

def flatten(l, ltypes=(list, tuple)):
    """Helper routine to flatten arrays"""
    i = 0
    # len(l) is eval'd each time --
    # no need for maxint with
    # try...except, just run to
    # the dynamic size of the list
    while (i < len(l)):
        while (isinstance(l[i], ltypes)):
            l[i:i+1] = list(l[i])
        i += 1
    return l

def filecheck(filename):
    """Check if filename exists and return filesize"""

    if filename.find("root://") != -1:
        # special check for EOS files
        cmd =  "%s %s// file info %s" % (config['PathToEOSBinary'], "/".join(filename.split("/")[:3]), '/'.join(filename.split("/")[3:]))
        rc, out = commands.getstatusoutput(cmd)

        if rc != 0:
            logger.debug("Problem checking EOS file '%s'. Ouptut: %s" % (filename, out))
            return -1

        # Assume first line is:
        # File: '/eos/atlas/user/m/mslater/tests/1.0/AnalysisSkeleton.aan.root'  Size: 1461388
        toks = out.split("\n")[0].split(":")
        if len(toks) != 3:
            logger.warning("Unexpected EOS output '%s'" % (out.split("\n")[0]))
            return 1

        fsize = int(toks[2])
        return fsize
    else:
        try:
            open(filename)
            fsize = os.stat(filename).st_size
        except IOError:
            logger.debug("File %s not found", filename)
            return -1
        
        if (fsize>0):
            return fsize

def expand(name):
    '''
    Expand file names according to ATLAS conventions
   
    Append the approriate LFC path

    Digits within square brackets are expanded
    lfn:rome.004100.recov10.T1_McAtNLO_top._[00001-10,00011,00014-20].AOD.pool.root
      
    '''
   
    n = name.split('.')
             
    if n[0] == 'rome':
        name = '/grid/atlas/datafiles/rome/%s/%s/%s' % (n[2],'.'.join(n[0:4]),name)
    elif n[0] in ['mc11','csc11']:
        if n[3] == 'recon':
            name = '/grid/atlas/datafiles/%s/recon/%s/%s' % (n[0],'.'.join(n[0:6]),name)
        else:
            name = '/grid/atlas/datafiles/%s/%s/%s/%s' % (n[0],n[3],'.'.join(n[0:5]),name)

    match1 = re.match('^(\S*)\[([0-9\-,]+)\](.*)$',name)
    if not match1: return [ name ] 
   
    names = []
    re_comma = re.compile('(\d+)-(\d+)')
    for part in match1.group(2).split(','):
        match2 = re_comma.match(part)
        if not match2:
            names += [ match1.group(1)+part+match1.group(3) ]
            continue
        l = max(len(match2.group(1)),len(match2.group(2)))
        i1 = int(match2.group(1))
        i2 = int(match2.group(2))
        names += [ match1.group(1) + '%0*d' % (l,i) + match1.group(3) for i in range(i1,i2+1) ]
      
    return names  

from commands import getstatusoutput    
import threading
from Ganga.Core import GangaThread

class Download:
    """Helper class for background download of files stored on remote SEs"""
    def __init__(self):
        super(Download, self).__init__()

    lfns = []
    prefix_hack = ''
    #rootfile = []
    lock = threading.RLock()

    class download_lcglr(GangaThread.GangaThread):
        def __init__(self, cmd):
            self.cmd = cmd
            GangaThread.GangaThread.__init__(self,'download_lcglr')

        def run(self):
            gridshell = getShell(middleware=Download.prefix_hack)

            gridshell.env['LFC_HOST'] = config['ATLASOutputDatasetLFC']
            gridshell.env['LCG_CATALOG_TYPE'] = 'lfc'
                         
            rc, out, m = gridshell.cmd1(self.cmd,allowed_exit=[0,255])
            #rc, out = getstatusoutput(self.cmd)
            if (rc==0):
                logger.debug("lcglr: %s", self.cmd)
                Download.lock.acquire()
                Download.lfns.append(out.strip())
                Download.lock.release()
            else:
                logger.error("Error occured during %s %s", self.cmd, out)

    class download_lcgcp(GangaThread.GangaThread):
        def __init__(self, cmd, ifile, pfn):
            self.cmd = cmd
            self.ifile = ifile
            self.pfn = pfn
            GangaThread.GangaThread.__init__(self,'download_lcgcp')
            
        def run(self):
            gridshell = getShell(middleware=Download.prefix_hack)
            gridshell.env['LFC_HOST'] = config['ATLASOutputDatasetLFC']
            gridshell.env['LCG_CATALOG_TYPE'] = 'lfc'
            rc, out, m = gridshell.cmd1(self.cmd,allowed_exit=[0,255])
            #rc, out = getstatusoutput(self.cmd)
            if (rc==0):
                logger.debug("lcg-cp finished: %s", self.cmd)
                logger.info("lcg-cp of %s finished", self.pfn)
                #Download.lock.acquire()
                #Download.rootfile[self.ifile].append(self.pfn)
                #Download.lock.release()
            else:
                logger.error("Error occured during %s %s", self.cmd, out)
    
    class download_dq2(GangaThread.GangaThread):
        def __init__(self, cmd):
            self.cmd = cmd
            GangaThread.GangaThread.__init__(self,'download_dq2')
            
        def run(self):
            gridshell = getShell(middleware=Download.prefix_hack)
            gridshell.env['DQ2_URL_SERVER']=configDQ2['DQ2_URL_SERVER']
            gridshell.env['DQ2_URL_SERVER_SSL']=configDQ2['DQ2_URL_SERVER_SSL']
            gridshell.env['DQ2_LOCAL_ID']=''

            ## Don't set up from the included DQ2 package as this will fail, either because
            # of python version (2.5+ required) or LFC python bindings missing
            
            #import GangaAtlas.PACKAGE
            #try:
            #    pythonpath=GangaAtlas.PACKAGE.setup.getPackagePath2('DQ2Clients','PYTHONPATH',force=False)
            #except:
            #    pythonpath = ''
            #gridshell.env['PYTHONPATH'] = gridshell.env['PYTHONPATH']+':'+pythonpath

            ## exclude the Ganga-owned external package for LFC python binding
            pythonpaths = []
            for path in gridshell.env['PYTHONPATH'].split(':'):
                if not re.match('.*\/external\/lfc\/.*', path):
                    pythonpaths.append(path)
            gridshell.env['PYTHONPATH'] = ':'.join(pythonpaths)
            
            ## exclude any rubbish from Athena
            ld_lib_paths = []
            for path in gridshell.env['LD_LIBRARY_PATH'].split(':'):
                if not re.match('.*\/external\/lfc\/.*', path) and not re.match('.*\/sw\/lcg\/external\/.*', path):
                    ld_lib_paths.append(path)
            gridshell.env['LD_LIBRARY_PATH'] = ':'.join(ld_lib_paths)
            
            paths = []
            for path in gridshell.env['PATH'].split(':'):
                if not re.match('.*\/external\/lfc\/.*', path) and not re.match('.*\/sw\/lcg\/external\/.*', path):
                    paths.append(path)
            gridshell.env['PATH'] = ':'.join(paths)
            
            rc, out, m = gridshell.cmd1("source " + configDQ2['setupScript'] + " && " + self.cmd,allowed_exit=[0,255])

            if (rc==0):
                logger.debug("dq2-get finished: %s", self.cmd)
                logger.debug("dq2-get output: %s %s %s"%(rc,out,m) )
                logger.warning("dq2-get finished")
            else:
                logger.error("Error occured during %s %s", self.cmd, out)


class ATLASDataset(Dataset):
    """ATLAS Datasets is a list of LFNs"""
    
    _schema = Schema(Version(1,1), {
        'lfn': SimpleItem(defvalue = [], typelist=['str'], sequence=1, doc='List of input file lfns'),
        'lfc': SimpleItem(defvalue='',sequence=0, doc='LFC Catalog address')
        
        })
    
    _category = 'datasets'
    _name = 'ATLASDataset'

    _exportmethods = ['get_dataset']

    _GUIPrefs = [ { 'attribute' : 'lfn',  'widget' : 'String_List' },
                  { 'attribute' : 'lfc',  'widget' : 'String' } ]

    def __init__(self):
        super(ATLASDataset, self).__init__()

    def filenames(self):
        """Get filenames by expanding the given names"""
       
        files=[]
        if self.lfn:
            for lfn in self.lfn:
                if self.lfc=='':
                    files+=expand(lfn)
                else:
                    files.append(lfn)
      
        return files

    def get_dataset(app):
        """Retrieve the file names starting from an application object"""
        job=app._getRoot()
        if not job:
            logger.warning('Application object is not associated to a job.')
            return []
         
#       jobs without inputdata are allowed
         
        if not job.inputdata: return []
      
        if not job.inputdata._name == 'ATLASDataset':
            logger.warning('Dataset is not of type ATLASDataset.')
            return []

        return job.inputdata.filenames()

    @staticmethod
    def get_filenames(app):
        """Retrieve the file names starting from an application object"""
      
        job=app._getRoot()
        if not job:
            logger.warning('Application object is not associated to a job.')
            return []
         
#       jobs without inputdata are allowed
         
        if not job.inputdata: return []
      
        if not job.inputdata._name == 'ATLASDataset':
            logger.warning('Dataset is not of type ATLASDataset.')
            return []

        return job.inputdata.filenames()


class ATLASTier3Dataset(Dataset):
    """ATLASTier3Dataset is a list of PFNs"""
    
    _schema = Schema(Version(1,0), {
        'names': SimpleItem(defvalue = [], typelist=['str'], sequence=1, doc='List of input file Physical File Names'),
        'pfnListFile': FileItem(doc='A text file containing a newline-separated list of Physical File Names'),
        })
    
    _category = 'datasets'
    _name = 'ATLASTier3Dataset'

    _exportmethods = []

    _GUIPrefs = [ { 'attribute' : 'names',  'widget' : 'String_List' },
                  { 'attribute' : 'pfnListFile',  'widget' : 'File' }
                 ]

    def __init__(self):
        super(ATLASTier3Dataset, self).__init__()

class ATLASLocalDataset(Dataset):
    """ATLAS Datasets is a list of local files"""
    
    _schema = Schema(Version(1,0), {
        'names': SimpleItem(defvalue = [], typelist=['str'], sequence=1,doc='List of input files with full path'),
        'use_poolfilecatalog_failover' : SimpleItem(defvalue = False, doc = 'Use pool_insertFileToCatalog per single file if bulk insert fails'),
        'create_poolfilecatalog' : SimpleItem(defvalue = False, doc = 'Try to add these files to the PoolFileCatalog'),
    })
    
    _category = 'datasets'
    _name = 'ATLASLocalDataset'

    _exportmethods = ['get_dataset', 'get_dataset_filenames', 'get_dataset_from_list' ]

    _GUIPrefs = [ { 'attribute' : 'names',  'widget' : 'String_List' } ]
   
    def __init__(self):
        super(ATLASLocalDataset, self).__init__()
        
    def get_dataset_from_list(self,list_file,no_dir_check = False):
       """Get the dataset files as listed in a text file"""

       logger.info('Reading list file %s ...',list_file)

       if not os.path.exists(list_file):
           logger.error('File %s does not exist',list_file)
           return

       f = open( list_file )
       for ln in f.readlines():

           # if no_dir_check then just copy the list of files
           if no_dir_check:
               self.names.append(ln.strip())
               continue

           # split the directory from the file and call get_dataset
           if os.path.isdir(ln.strip()):
               self.get_dataset( ln.strip() )
           else:
               self.get_dataset( os.path.dirname(ln.strip()), os.path.basename( ln.strip() ) )
           
    def get_dataset(self,directory,filter=None):
       """Get the actual files of a dataset"""
      
       logger.info('Reading %s ...',directory)


       if not os.path.isdir(directory):
           logger.error('Path %s is no directory',directory)
           return

       directory = os.path.abspath(directory)
       if filter:
           new_names = [ os.path.join(directory,name) for name in fnmatch.filter(os.listdir(directory),filter) ]
       else:
           new_names = [ os.path.join(directory,name) for name in os.listdir(directory) ]

       self.names.extend( new_names )

       self._setDirty()

    def get_dataset_filenames(self):
        """Get filenames"""
        return self.names

    @staticmethod
    def get_filenames(app):
        """Retrieve the file names starting from an application object"""
      
        job=app._getRoot()
        if not job:
            logger.warning('Application object is not associated to a job.')
            return []
         
#       jobs without inputdata are allowed
         
        if not job.inputdata: return []
      
        if not job.inputdata._name == 'ATLASLocalDataset':
            logger.warning('Dataset is not of type ATLASLocalDataset.')
            return []

        return job.inputdata.names


class ATLASOutputDataset(Dataset):
    """Generic ATLAS Dataset for a list of different outputfiles"""
    
    _schema = Schema(Version(1,0), {
        'outputdata'     : SimpleItem(defvalue = [], typelist=['str'], sequence=1, doc='Output files to be returned via SE'), 
        'output'         : SimpleItem(defvalue = [], typelist=['str'], sequence = 1, protected=1, doc = 'Output information automatically filled by the job'),
        'location'       : SimpleItem(defvalue='',doc='SE output path location'),
        'local_location' : SimpleItem(defvalue='',doc='Local output path location')
        })
    
    _category = 'datasets'
    _name = 'ATLASOutputDataset'

    _exportmethods = [ 'retrieve', 'fill' ]

    _GUIPrefs = [ { 'attribute' : 'outputdata',     'widget' : 'String_List' },
                  { 'attribute' : 'output',         'widget' : 'String_List' },
                  { 'attribute' : 'location',       'widget' : 'String' },
                  { 'attribute' : 'local_location', 'widget' : 'File' } ]

    def __init__(self):
        super(ATLASOutputDataset, self).__init__()

    def fill(self, type=None, name=None, **options ):
        """Determine outputdata and outputsandbox locations of finished jobs
        and fill output variable"""
        from Ganga.GPIDev.Lib.Job import Job
        job = self._getParent()

#       Determine local output path to store files
        if job.outputdata.location and ((job.backend._name == 'Local') or (job.backend._name == 'LSF') or (job.backend._name == 'PBS') or (job.backend._name == 'SGE') or (job.backend._name == 'Condor')):
            outputlocation =  expandfilename(job.outputdata.location)
            # update the local_location variable to point to the location
            job.outputdata.local_location = outputlocation
        elif job.outputdata.local_location:
            outputlocation = expandfilename(job.outputdata.local_location)
        else:
            try:
                tmpdir = os.environ['TMPDIR']
            except:
                tmpdir = '/tmp/'
            outputlocation = tmpdir

#       Output files on SE
        outputfiles = job.outputdata.outputdata
        
#       Search output_guid files from LCG jobs in outputsandbox
        jobguids = []

        if (job.backend._name == 'LCG' ):
            pfn = job.outputdir + "output_guids"
            fsize = filecheck(pfn)
            if (fsize>0):
                jobguids.append(pfn)
                logger.debug('jobsguids: %s', jobguids)
            
#       Get guids from output_guid files
            for ijobguids in jobguids: 
                f = open(ijobguids)
                tempguids =  [ line.strip() for line in f ]
                if not self.output:
                    self.output = self.output + tempguids

#       Local host execution
        if (job.backend._name == 'Local' or \
            job.backend._name == 'LSF' or \
            job.backend._name == 'PBS' or \
            job.backend._name == 'SGE' or \
            job.backend._name == 'Condor'):
            for file in outputfiles:

                pfn = outputlocation+"/"+file
                fsize = filecheck(pfn)
                if (fsize>0):
                    self.output.append(pfn)

#       Output files in the sandbox 
        outputsandboxfiles = job.outputsandbox
        for file in outputsandboxfiles:
            pfn = job.outputdir+"/"+file
            fsize = filecheck(pfn)
            if (fsize>0):
                self.output.append(pfn)

#       Master job finish
        if not job.master and job.subjobs:
            for subjob in job.subjobs:
                self.output.append(subjob.outputdata.output)


    def retrieve(self, type=None, name=None, **options ):
        """Retieve files listed in outputdata and registered in output from
        remote SE to local filesystem in background thread"""
        from Ganga.GPIDev.Lib.Job import Job
        import os
        
        job = self._getParent()

        try:
            logger.info("ATLASDataset.retrieve() called.")
            logger.debug('job.id: %d, Job.subjobs: %d',job.id,len(job.subjobs))
            logger.debug('job.outputdir: %s',job.outputdir)
            logger.debug('job.outputsandbox: %s',job.outputsandbox)
            logger.debug('job.outputdata.outputsandbox: %s',job.outputdata)
            logger.debug('job.outputdata.outputdata: %s',job.outputdata.outputdata)
            logger.debug('job.outputdata.output: %s',job.outputdata.output)
            logger.debug('job.outputdata.location: %s',job.outputdata.location)
            logger.debug('job.outputdata.local_location: %s',job.outputdata.local_location)
            
        except AttributeError:
            logger.error('job.outputdata error')
            return 1

        if job.backend._name == 'LCG':
            Download.prefix_hack = job.backend.middleware
            
        else:
            Download.prefix_hack = 'EDG'

        local_location = options.get('local_location')

#       Determine local output path to store files
        outputlocation = ''
        if job.outputdata.location and (job.backend._name == 'Local'):
            outputlocation = expandfilename(job.outputdata.location)
        elif local_location:
            outputlocation = expandfilename(local_location) 
        elif job.outputdata.local_location:
            outputlocation = expandfilename(job.outputdata.local_location)
        else:
            # User job repository location
            #outputlocation = job.outputdir+'/../'            
            pass
            
#       Output files 
        outputfiles = job.outputdata.outputdata
        
#       Search output_guid files from LCG jobs in outputsandbox
        if (job.backend._name == 'LCG' ):
            threads = []
            outputnew = flatten(self.output)
             
            for iguid in outputnew:
                cmd = None
                if iguid[0:5]=="guid:":
                    cmd = "lcg-lr --vo atlas "+iguid
                if cmd:
                    thread = Download.download_lcglr(cmd)
                    thread.setDaemon(True)
                    thread.start()
                    threads.append(thread)
                
            for thread in threads:
                thread.join()

            logger.debug('output %s', outputnew)
            logger.debug('lfns %s', Download.lfns)
#        Download the lfns of every subjob to local outputlocation subdirectory
            threads = []
            if (len(Download.lfns)!=0):
                logger.info("Please be patient - background downloading files from remote SE.")
                # Master job retrieve
                pfnid = []
                if not job.master and job.subjobs and outputlocation=='':
                    for subjob in job.subjobs:
                        id = "%d/%d" % (job._getRoot().id,subjob.id)
                        pfnid.append(id)
                elif not job.master and job.subjobs and outputlocation != job.outputdir:
                    for subjob in job.subjobs:
                        id = "%d/%d" % (job._getRoot().id, subjob.id)
                        pfnid.append(id)
                        try:
                            os.makedirs(outputlocation+"/"+id)
                        except OSError:
                            pass
                else:
                    if outputlocation=='':
                        id = "%d/%d" % (job._getRoot().id,job.id)
                        pfnid.append(id)
                    elif outputlocation != job.outputdir:
                        if job._getRoot().subjobs:
                            id = "%d/%d" % (job._getRoot().id, job.id)
                        else:
                            id = "%d" % job.id
                        try:
                            os.makedirs(outputlocation+"/"+id)
                        except OSError:
                            pass

                for ilfn in range(len(Download.lfns)):
                    for ifile in xrange(len(outputfiles)):
                        # Find correct jobid
                        for ipfnid in pfnid:
                            ipfnid = re.sub('/','.',ipfnid)
                            if (Download.lfns[ilfn]).find(ipfnid)>0:
                                
                                ipfnid = re.sub('\.','/',ipfnid)
                                id = ipfnid

                        print id, pfnid, Download.lfns
                        
                        if (Download.lfns[ilfn]).find(outputfiles[ifile])>0:
                            if outputlocation == '':
                                if job.subjobs:
                                    tempid = int(id.split('/')[1])
                                    pfn = job.subjobs[tempid].outputdir+outputfiles[ifile]
                                else:
                                    pfn = job.outputdir+outputfiles[ifile]
                                    
                            elif outputlocation != job.outputdir:
                                pfn = outputlocation+"/"+id+"/"+outputfiles[ifile]
                            else:
                                pfn = outputlocation+outputfiles[ifile]
                            
                            logger.info("Downloading %s", pfn)
                            cmd = "lcg-cp --vo atlas "+Download.lfns[ilfn]+" file:"+pfn
                            thread = Download.download_lcgcp(cmd,ifile,pfn)
                            thread.setDaemon(True)
                            thread.start()
                            threads.append(thread)

#                for thread in threads:
#                    thread.join()
#                logger.info("Downloading finished.")

            Download.lfns = []


#        logger.debug('Download.rootfile %s', Download.rootfile)


#$Log: not supported by cvs2svn $
#Revision 1.7  2009/04/24 08:36:21  dvanders
##49529: provide name for GangaThread
#
#Revision 1.6  2009/03/19 15:46:17  dvanders
#Thread->GangaThread
#
#Revision 1.5  2009/01/29 15:46:50  mslater
#Removed unnecessary print statement in Download class
#
#Revision 1.4  2008/08/19 13:15:32  elmsheus
#Fix bug #39010, Add ATLASLocalDataset support for list of datasets
#
#Revision 1.3  2008/07/28 15:02:30  elmsheus
#Fix for bug #35256
#
#Revision 1.2  2008/07/28 14:27:34  elmsheus
#* Upgrade to DQ2Clients 0.1.17 and DQ2 API
#* Add full support for DQ2 container datasets in DQ2Dataset
#* Change in DQ2OutputDataset.retrieve(): use dq2-get
#* Fix bug #39286: Athena().atlas_environment omits type_list
#
#Revision 1.1  2008/07/17 16:41:18  moscicki
#migration of 5.0.2 to HEAD
#
#the doc and release/tools have been taken from HEAD
#
#Revision 1.30.2.5  2008/05/12 15:55:38  elmsheus
#Fix small typo
#
#Revision 1.30.2.4  2008/05/12 09:07:31  elmsheus
#Add SGE output support
#
#Revision 1.30.2.3  2008/03/07 20:26:22  elmsheus
#* Apply Ganga-5-0-restructure-config-branch patch
#* Move GangaAtlas-4-15 tag to GangaAtlas-5-0-branch
#
#Revision 1.30.2.2  2008/02/18 11:03:22  elmsheus
#Copy GangaAtlas-4-13 to GangaAtlas-5-0-branch and config updates
#
#Revision 1.31  2008/03/03 14:11:40  elmsheus
#Athena:
#* Fix problem in requirements file creation - directories on lowest
#  level are now included by cmt
#* Fix DQ2JobSplitter problem with unused datasets - add exception
#* Fix problem in athena-local.sh - add ServiceMgr correctly
#* Add missing environment variables in DQ2OutputDataset.retrieve()
#* Add python32 detection in ganga-stage-in-out-dq2.py if dq2_get is
#  used
#
#Revision 1.30  2007/09/28 12:31:35  elmsheus
#Add local_location option to retrieve method
#
#Revision 1.29  2007/09/24 15:13:47  elmsheus
#* Change output path in ATLASOutputDataset.retrieve() methode from
#  jobid.subjobid to jobid/subjobid
#* Fix ATLASOutputDataset merging to actually find previously downloaded
#  files
#
#Revision 1.28  2007/09/24 08:42:10  elmsheus
#Apply patches to migrate to Ganga.Utility.GridShell
#
#Revision 1.27  2007/05/23 13:28:31  elmsheus
#Add ATLASOutputDatasetLFC config variable to set LFC host for ATLASOutputDataset
#
#Revision 1.26  2007/04/02 08:07:25  elmsheus
#* Fix directory scanning procedure in Athena.prepare()
#* Fix GUIPrefs problems
#
#Revision 1.25  2007/03/24 15:36:26  liko
#Fix get_dataset in ATLASLocalDataset
#
#Revision 1.24  2007/03/21 15:11:29  elmsheus
#Add GUIPrefs
#
#Revision 1.23  2007/02/22 12:55:41  elmsheus
#Fix output path and use gridShell
#
#Revision 1.22  2007/02/21 08:09:04  elmsheus
#Small fixes
#
#Revision 1.21  2007/02/20 17:39:40  elmsheus
#Introduce new functionality in ATLASDataset: ganga-stagein-lfc.py
#specify list of lfns and lfc host
#direct access to inputdata files on worker node
#
#Revision 1.20  2007/02/12 15:31:42  elmsheus
#Port 4.2.8 changes to head
#Fix job.splitter in Athena*RTHandler
#
#Revision 1.19  2006/11/27 12:18:02  elmsheus
#Fix CVS merging errors
#
#Revision 1.18  2006/11/24 15:39:13  elmsheus
#Small fixes
#
#Revision 1.17  2006/11/24 13:32:37  elmsheus
#Merge changes from Ganga-4-2-2-bugfix-branch to the trunk
#Add Frederics changes and improvement for AthenaMC
#
#Revision 1.16.2.4  2006/11/24 15:15:52  elmsheus
#Fix prefix_hack
#
#Revision 1.16.2.3  2006/11/22 14:20:52  elmsheus
#* introduce prefix_hack to lcg-cp/lr calls in
#  ATLASOutputDataset.retrieve()
#* fixed double downloading feature in
#  ATLASOutputDataset.retrieve()
#* move download location for ATLASOutputDataset.retrieve()
#  to job.outputdir from temp directory if local_location is not given
#* Print out clear error message if cmt parsing fails in Athena.py
#* Migrate to GridProxy library in Athena*RTHandler.py
#* Changes in output renaming schema for DQ2OutputDataset files
#
#* Fix proxy name bug in AthenaMCLCGRTHandler.py
#* Fix path problem in wrapper.sh
#
#Revision 1.16.2.2  2006/11/07 09:41:09  elmsheus
#Enable outputdata.retrieve() also for master job
#Add 'addAANT' root tuple merging
#
#Revision 1.16.2.1  2006/10/27 15:31:57  elmsheus
#Fix output location issue
#
#Revision 1.16  2006/10/23 07:21:27  elmsheus
#Fix AthenaLocaDataset bug
#
#Revision 1.15  2006/10/16 11:59:25  elmsheus
#Change ATLASLocalDataset get_dataset
#
#Revision 1.14  2006/09/08 16:11:44  elmsheus
#Expand SimpleItem directory variables with expandfilenames
#
#Revision 1.13  2006/09/08 14:40:46  elmsheus
#Fix ouput bug for LSF, PBS jobs
#
#Revision 1.12  2006/08/09 16:47:12  elmsheus
#Introduction of DQ2OutputDataset, fix minor bugs
#
#Revision 1.11  2006/07/09 08:41:01  elmsheus
#ATLASOutputDataset introduction, DQ2 updates, Splitter and Merger code clean-up, and more
#
#Revision 1.10  2006/05/27 10:18:01  elmsheus
#Fix ATLASLocalDataset processing problem
#
#Revision 1.9  2006/05/09 13:45:29  elmsheus
#Introduction of
# Athena job splitting based on number of subjobs
# DQ2Dataset and DQ2 file download
# AthenaLocalDataset
#
#Revision 1.8  2006/03/19 01:35:28  liko
#repair datasets
#
#Revision 1.7  2006/03/17 00:55:53  liko
#Fixes in ATLASDataset
#
#Revision 1.6  2006/03/15 16:30:07  liko
#Remove DIAL & ADA
#
#Revision 1.5  2006/03/15 15:49:14  liko
#Small fixes in ATLASCastorDataset
#
#Revision 1.4  2005/10/20 09:47:25  karl
#KH: minor correction
#
#Revision 1.3  2005/09/21 23:37:29  liko
#Bugfixes and support for Simulation
#
#Revision 1.2  2005/09/06 11:37:14  liko
#Mainly the Athena handler
#
